from typing import List, Dict, Any, Set
import math
from scipy.special import rel_entr
from csle_common.dao.system_identification.gaussian_mixture_conditional import GaussianMixtureConditional
from csle_common.dao.system_identification.system_model import SystemModel
from csle_common.dao.system_identification.system_model_type import SystemModelType


class GaussianMixtureSystemModel(SystemModel):
    """
    A system model (list of conditional distributions) made up of Gaussian Mixtures
    """

    def __init__(self, emulation_env_name: str, emulation_statistic_id: int,
                 conditional_metric_distributions: List[List[GaussianMixtureConditional]], descr: str):
        """
        Initializes the object

        :param emulation: the emulation that this system model is for
        :param emulation_statistic_id: the emulation statistic that this model was built from
        :param conditional_metric_distributions: the list of conditional distributions
        :param descr: description of the model
        """
        super(GaussianMixtureSystemModel, self).__init__(descr=descr, model_type=SystemModelType.GAUSSIAN_MIXTURE)
        self.conditional_metric_distributions = conditional_metric_distributions
        complete_sample_space: Set[int] = set()
        for conds in self.conditional_metric_distributions:
            for cond in conds:
                complete_sample_space = complete_sample_space.union(cond.sample_space)
        for conds in self.conditional_metric_distributions:
            for cond in conds:
                cond.sample_space = list(complete_sample_space)
                cond.generate_distributions()
        self.emulation_env_name = emulation_env_name
        self.emulation_statistic_id = emulation_statistic_id
        self.id = -1
        self.conditionals_kl_divergences: Dict[str, Dict[str, Dict[str, float]]] = {}
        self.compute_kl_divergences()

    def compute_kl_divergences(self) -> None:
        """
        Computes the KL-divergences betwen different conditional distributions

        :return: None
        """
        for metric_distributions_condition_1 in self.conditional_metric_distributions:
            self.conditionals_kl_divergences[metric_distributions_condition_1[0].conditional_name] = {}
            for metric_distributions_condition_2 in self.conditional_metric_distributions:
                self.conditionals_kl_divergences[metric_distributions_condition_1[0].conditional_name][
                    metric_distributions_condition_2[0].conditional_name] = {}
                for i, metric_dist in enumerate(metric_distributions_condition_1):
                    self.conditionals_kl_divergences[metric_distributions_condition_1[0].conditional_name][
                        metric_distributions_condition_2[0].conditional_name][metric_dist.metric_name] = float(
                        round(sum(rel_entr(metric_dist.combined_distribution,
                                           metric_distributions_condition_2[i].combined_distribution)), 3))
                    if math.isinf(
                            self.conditionals_kl_divergences[metric_distributions_condition_1[0].conditional_name][
                                metric_distributions_condition_2[0].conditional_name][metric_dist.metric_name]):
                        self.conditionals_kl_divergences[metric_distributions_condition_1[0].conditional_name][
                            metric_distributions_condition_2[0].conditional_name][metric_dist.metric_name] = math.inf

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "GaussianMixtureSystemModel":
        """
        Converts a dict representation of the DTO into an instance

        :param d: the dict to convert
        :return: the converted instance
        """
        dto = GaussianMixtureSystemModel(
            emulation_env_name=d["emulation_env_name"],
            emulation_statistic_id=d["emulation_statistic_id"],
            conditional_metric_distributions=list(map(lambda x:
                                                      list(map(lambda y:
                                                               GaussianMixtureConditional.from_dict(y), x)),
                                                      d["conditional_metric_distributions"])),
            descr=d["descr"]
        )
        if "id" in d:
            dto.id = d["id"]
        return dto

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the DTO
        """
        d: Dict[str, Any] = {}
        d["conditional_metric_distributions"] = list(map(lambda x: list(map(lambda y: y.to_dict(), x)),
                                                         self.conditional_metric_distributions))
        d["emulation_env_name"] = self.emulation_env_name
        d["emulation_statistic_id"] = self.emulation_statistic_id
        d["descr"] = self.descr
        d["id"] = self.id
        d["conditionals_kl_divergences"] = self.conditionals_kl_divergences
        d["model_type"] = self.model_type
        return d

    def __str__(self):
        """
        :return: a string representation of the DTO
        """
        return f"conditional_distributions: {self.conditional_metric_distributions}, " \
               f"emulation_env_name: {self.emulation_env_name}, " \
               f"emulation_statistic_id: {self.emulation_statistic_id}," \
               f"descr: {self.descr}, conditionals_kl_divergences: {self.conditionals_kl_divergences}, " \
               f"model_type: {self.model_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "GaussianMixtureSystemModel":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return GaussianMixtureSystemModel.from_dict(json.loads(json_str))

    def copy(self) -> "GaussianMixtureSystemModel":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())



import math
from typing import Dict, Any, List
import numpy as np
from scipy.special import rel_entr
import csle_common.constants.constants as constants
import csle_collector.constants.constants as collector_constants
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_base.json_serializable import JSONSerializable


class EmulationStatistics(JSONSerializable):
    """
    DTO representing delta-statistics measured from teh emulation
    """

    def __init__(self, emulation_name: str, descr: str = ""):
        """
        Initializes the statistics DTO

        :param emulation_name: the name of the emulation that the statistics is linked to
        :param descr: a free text description of the statistics
        """
        self.emulation_name = emulation_name
        self.descr = descr
        self.initial_distributions_counts = self.initialize_counters(
            d={}, agg_labels=collector_constants.KAFKA_CONFIG.ALL_INITIAL_AGG_LABELS)
        self.conditionals_counts = {}
        self.conditionals_counts[constants.SYSTEM_IDENTIFICATION.INTRUSION_CONDITIONAL] = \
            EmulationStatistics.initialize_counters(d={},
                                                    agg_labels=collector_constants.KAFKA_CONFIG.ALL_DELTA_AGG_LABELS)
        self.conditionals_counts[constants.SYSTEM_IDENTIFICATION.NO_INTRUSION_CONDITIONAL] = \
            EmulationStatistics.initialize_counters(d={},
                                                    agg_labels=collector_constants.KAFKA_CONFIG.ALL_DELTA_AGG_LABELS)
        self.id = -1

        self.means: Dict[str, Any] = {}
        self.stds: Dict[str, Any] = {}
        self.mins: Dict[str, Any] = {}
        self.maxs: Dict[str, Any] = {}
        self.conditionals_probs: Dict[str, Any] = {}
        self.initial_distributions_probs: Dict[str, Any] = {}
        self.initial_means: Dict[str, Any] = {}
        self.initial_stds: Dict[str, Any] = {}
        self.initial_mins: Dict[str, Any] = {}
        self.initial_maxs: Dict[str, Any] = {}
        self.conditionals_kl_divergences: Dict[str, Any] = {}
        self.num_metrics = 0
        self.num_measurements = 0
        self.num_conditions = 0
        self.conditions: List[Any] = []
        self.metrics: List[Any] = []

    @staticmethod
    def initialize_counters(d: Dict[str, Dict[int, int]], agg_labels: List[str]) \
            -> Dict[str, Dict[int, int]]:
        """
        Initializes counters for a given dict

        :param d: the dict to initialize
        :param agg_labels: the labels
        :return: the updated dict
        """
        for label in agg_labels:
            d[label] = {}

        return d

    @staticmethod
    def initialize_machine_counters(d: Dict[str, Dict[int, int]], s: EmulationEnvState, labels: List[str]) \
            -> Dict[str, Dict[int, int]]:
        """
        Initializes counters for a given dict

        :param d: the dict to initialize
        :param s: the state with the list of machines
        :param labels: the labels to initialize
        :return: the initialized dict
        """
        labels = collector_constants.KAFKA_CONFIG.ALL_INITIAL_MACHINE_LABELS
        if s.defender_obs_state is None:
            raise ValueError("EmulationDefenderObservationState is None")
        for label in labels:
            for machine in s.defender_obs_state.machines:
                lbl = f"{label}_{machine.ips[0]}"
                d[lbl] = {}
        return d

    def initialize_machines(self, s: EmulationEnvState) -> None:
        """
        Initializes counters for a given dict

        :param d: the dict to initialize
        :param s: the state with the list of machines
        :return: the initialized dict
        """
        self.initial_distributions_counts = EmulationStatistics.initialize_machine_counters(
            d=self.initial_distributions_counts, s=s,
            labels=collector_constants.KAFKA_CONFIG.ALL_INITIAL_MACHINE_LABELS)
        self.conditionals_counts[constants.SYSTEM_IDENTIFICATION.INTRUSION_CONDITIONAL] = \
            EmulationStatistics.initialize_machine_counters(
                s=s,
                d=self.conditionals_counts[constants.SYSTEM_IDENTIFICATION.INTRUSION_CONDITIONAL],
                labels=collector_constants.KAFKA_CONFIG.ALL_DELTA_MACHINE_LABELS)
        self.conditionals_counts[constants.SYSTEM_IDENTIFICATION.NO_INTRUSION_CONDITIONAL] = \
            EmulationStatistics.initialize_machine_counters(
                s=s,
                d=self.conditionals_counts[constants.SYSTEM_IDENTIFICATION.NO_INTRUSION_CONDITIONAL],
                labels=collector_constants.KAFKA_CONFIG.ALL_DELTA_MACHINE_LABELS)

    def update_counters(self, d: Dict[str, Any], s: EmulationEnvState, s_prime: EmulationEnvState) -> None:
        """
        Updates the delta counters for a specific dict based on a state transition s->s'

        :param d: the dict to update
        :param s: the current state
        :param s_prime: the new state
        :return: None
        """

        # Snort alerts
        if s_prime.defender_obs_state is None or s.defender_obs_state is None:
            raise ValueError("EmulationDefenderObservationState is None")
        snort_alert_deltas, snort_alert_labels = s.defender_obs_state.avg_snort_ids_alert_counters.get_deltas(
            s_prime.defender_obs_state.avg_snort_ids_alert_counters)
        for i in range(len(snort_alert_deltas)):
            if snort_alert_deltas[i] in d[snort_alert_labels[i]]:
                d[snort_alert_labels[i]][snort_alert_deltas[i]] += 1
            else:
                d[snort_alert_labels[i]][snort_alert_deltas[i]] = 1
        for machine in s.defender_obs_state.machines:
            s_prime_machine = s_prime.get_defender_machine(ip=machine.ips[0])
            if machine.snort_ids_ip_alert_counters is None or s_prime_machine is None:
                raise ValueError("Machine is None")
            snort_alert_deltas, snort_alert_labels = machine.snort_ids_ip_alert_counters.get_deltas(
                s_prime_machine.snort_ids_ip_alert_counters)
            for i in range(len(snort_alert_deltas)):
                lbl = f"{snort_alert_labels[i]}_{machine.ips[0]}"
                if snort_alert_deltas[i] in d[lbl]:
                    d[lbl][snort_alert_deltas[i]] += 1
                else:
                    d[lbl][snort_alert_deltas[i]] = 1

        # OSSEC alerts
        ossec_alert_deltas, ossec_alert_labels = s.defender_obs_state.avg_ossec_ids_alert_counters.get_deltas(
            s_prime.defender_obs_state.avg_ossec_ids_alert_counters)
        for i in range(len(ossec_alert_deltas)):
            if ossec_alert_deltas[i] in d[ossec_alert_labels[i]]:
                d[ossec_alert_labels[i]][ossec_alert_deltas[i]] += 1
            else:
                d[ossec_alert_labels[i]][ossec_alert_deltas[i]] = 1

        for machine in s.defender_obs_state.machines:
            s_prime_machine = s_prime.get_defender_machine(ip=machine.ips[0])
            if machine.ossec_ids_alert_counters is None or s_prime_machine is None:
                raise ValueError("Machine is None")
            ossec_alert_deltas, ossec_alert_labels = machine.ossec_ids_alert_counters.get_deltas(
                s_prime_machine.ossec_ids_alert_counters)
            for i in range(len(ossec_alert_deltas)):
                lbl = f"{ossec_alert_labels[i]}_{machine.ips[0]}"
                if ossec_alert_deltas[i] in d[lbl]:
                    d[lbl][ossec_alert_deltas[i]] += 1
                else:
                    d[lbl][ossec_alert_deltas[i]] = 1

        # Docker stats
        docker_stats_deltas, docker_stats_labels = s.defender_obs_state.avg_docker_stats.get_deltas(
            stats_prime=s_prime.defender_obs_state.avg_docker_stats)
        for i in range(len(docker_stats_deltas)):
            if docker_stats_deltas[i] in d[docker_stats_labels[i]]:
                d[docker_stats_labels[i]][docker_stats_deltas[i]] += 1
            else:
                d[docker_stats_labels[i]][docker_stats_deltas[i]] = 1

        for machine in s.defender_obs_state.machines:
            s_prime_machine = s_prime.get_defender_machine(ip=machine.ips[0])
            if machine.docker_stats is None or s_prime_machine is None:
                raise ValueError("Machine is None")
            docker_stats_deltas, docker_stats_labels = machine.docker_stats.get_deltas(
                stats_prime=s_prime_machine.docker_stats)
            for i in range(len(docker_stats_deltas)):
                lbl = f"{docker_stats_labels[i]}_{machine.ips[0]}"
                if docker_stats_deltas[i] in d[lbl]:
                    d[lbl][docker_stats_deltas[i]] += 1
                else:
                    d[lbl][docker_stats_deltas[i]] = 1

        # Client metrics
        client_population_metrics_deltas, client_population_metrics_labels = (
            s.defender_obs_state.avg_client_population_metrics.get_deltas(
                stats_prime=s_prime.defender_obs_state.avg_client_population_metrics))
        for i in range(len(client_population_metrics_deltas)):
            if client_population_metrics_deltas[i] in d[client_population_metrics_labels[i]]:
                d[client_population_metrics_labels[i]][client_population_metrics_deltas[i]] += 1
            else:
                d[client_population_metrics_labels[i]][client_population_metrics_deltas[i]] = 1

        # Host metrics
        if s.defender_obs_state is not None and s.defender_obs_state.avg_aggregated_host_metrics is not None:
            aggregated_host_metrics_deltas, aggregated_host_metrics_labels = \
                s.defender_obs_state.avg_aggregated_host_metrics.get_deltas(
                    stats_prime=s_prime.defender_obs_state.avg_aggregated_host_metrics)
        else:
            raise ValueError("Machine is None")
        for i in range(len(aggregated_host_metrics_deltas)):
            if aggregated_host_metrics_deltas[i] in d[aggregated_host_metrics_labels[i]]:
                d[aggregated_host_metrics_labels[i]][aggregated_host_metrics_deltas[i]] += 1
            else:
                d[aggregated_host_metrics_labels[i]][aggregated_host_metrics_deltas[i]] = 1

        for machine in s.defender_obs_state.machines:
            s_prime_machine = s_prime.get_defender_machine(ip=machine.ips[0])
            if machine.host_metrics is None or s_prime_machine is None:
                raise ValueError("Machine is None")
            host_metrics_deltas, host_metrics_labels = machine.host_metrics.get_deltas(
                stats_prime=s_prime_machine.host_metrics)
            for i in range(len(host_metrics_deltas)):
                lbl = f"{host_metrics_labels[i]}_{machine.ips[0]}"
                if host_metrics_deltas[i] in d[lbl]:
                    d[lbl][host_metrics_deltas[i]] += 1
                else:
                    d[lbl][host_metrics_deltas[i]] = 1

    def update_delta_statistics(self, s: EmulationEnvState, s_prime: EmulationEnvState, a1: EmulationDefenderAction,
                                a2: EmulationAttackerAction) -> None:
        """
        Updates the emulation statistics (delta counters) with a given transition (s, a1, a2) -> (s')

        :param s: the previous state
        :param s_prime: the new state
        :param a1: the defender action
        :param a2: the attacker action
        :return: None
        """

        # Intrusion vs No Intrusion Conditionals
        if a2.id == EmulationAttackerActionId.CONTINUE:
            self.update_counters(d=self.conditionals_counts[constants.SYSTEM_IDENTIFICATION.NO_INTRUSION_CONDITIONAL],
                                 s=s, s_prime=s_prime)
        else:
            self.update_counters(d=self.conditionals_counts[constants.SYSTEM_IDENTIFICATION.INTRUSION_CONDITIONAL],
                                 s=s, s_prime=s_prime)
        if s.attacker_obs_state is None:
            raise ValueError("EmulationAttackerObservationState is None")
        logged_in_ips = list(map(lambda x: "_".join(x.ips), filter(
            lambda x: x.logged_in and x.tools_installed and x.backdoor_installed and x.root,
            s.attacker_obs_state.machines)))
        # Action conditionals
        if f"A:{a2.name}_D:{a1.name}_M:{logged_in_ips}" not in self.conditionals_counts:
            self.conditionals_counts[f"A:{a2.name}_D:{a1.name}_M:{logged_in_ips}"] = \
                EmulationStatistics.initialize_counters(
                    d={}, agg_labels=collector_constants.KAFKA_CONFIG.ALL_DELTA_AGG_LABELS)
            self.conditionals_counts[f"A:{a2.name}_D:{a1.name}_M:{logged_in_ips}"] = \
                EmulationStatistics.initialize_machine_counters(
                    s=s,
                    d=self.conditionals_counts[f"A:{a2.name}_D:{a1.name}_M:{logged_in_ips}"],
                    labels=collector_constants.KAFKA_CONFIG.ALL_DELTA_MACHINE_LABELS)
        self.update_counters(
            d=self.conditionals_counts[f"A:{a2.name}_D:{a1.name}_M:{logged_in_ips}"], s=s, s_prime=s_prime)

    def update_initial_statistics(self, s: EmulationEnvState) -> None:
        """
        Updates the emulation statistics for the initial state

        :param s: the initial state
        :return: None
        """
        snort_alert_labels = collector_constants.KAFKA_CONFIG.SNORT_IDS_ALERTS_LABELS
        if s.defender_obs_state is None:
            raise ValueError("EmulationDefenderObservationState is None")
        for i in range(len(snort_alert_labels)):
            if 0 in self.initial_distributions_counts[snort_alert_labels[i]]:
                self.initial_distributions_counts[snort_alert_labels[i]][0] += 1
            else:
                self.initial_distributions_counts[snort_alert_labels[i]][0] = 1
            for machine in s.defender_obs_state.machines:
                lbl = f"{snort_alert_labels[i]}_{machine.ips[0]}"
                if 0 in self.initial_distributions_counts[lbl]:
                    self.initial_distributions_counts[lbl][0] += 1
                else:
                    self.initial_distributions_counts[lbl][0] = 1

        ossec_alert_labels = collector_constants.KAFKA_CONFIG.OSSEC_IDS_ALERTS_LABELS
        for i in range(len(ossec_alert_labels)):
            if 0 in self.initial_distributions_counts[ossec_alert_labels[i]]:
                self.initial_distributions_counts[ossec_alert_labels[i]][0] += 1
            else:
                self.initial_distributions_counts[ossec_alert_labels[i]][0] = 1
            for machine in s.defender_obs_state.machines:
                lbl = f"{ossec_alert_labels[i]}_{machine.ips[0]}"
                if 0 in self.initial_distributions_counts[lbl]:
                    self.initial_distributions_counts[lbl][0] += 1
                else:
                    self.initial_distributions_counts[lbl][0] = 1
        if s.defender_obs_state.docker_stats is None:
            raise ValueError("DockerStats is None")
        docker_stats_values, docker_stats_labels = s.defender_obs_state.docker_stats.get_values()
        for i in range(len(docker_stats_values)):
            if docker_stats_values[i] in self.initial_distributions_counts[docker_stats_labels[i]]:
                self.initial_distributions_counts[docker_stats_labels[i]][docker_stats_values[i]] += 1
            else:
                self.initial_distributions_counts[docker_stats_labels[i]][docker_stats_values[i]] = 1
        for machine in s.defender_obs_state.machines:
            if machine.docker_stats is None:
                raise ValueError("Dockerstats is None")
            m_values, m_labels = machine.docker_stats.get_values()
            for i in range(len(m_values)):
                lbl = f"{m_labels[i]}_{machine.ips[0]}"
                if m_values[i] in self.initial_distributions_counts[lbl]:
                    self.initial_distributions_counts[lbl][m_values[i]] += 1
                else:
                    self.initial_distributions_counts[lbl][m_values[i]] = 1
        if s.defender_obs_state.client_population_metrics is None:
            raise ValueError("ClientPopulationMetrics is None")
        client_population_metrics_values, client_population_metrics_labels = \
            s.defender_obs_state.client_population_metrics.get_values()
        for i in range(len(client_population_metrics_values)):
            if (client_population_metrics_values[i] in
                    self.initial_distributions_counts[client_population_metrics_labels[i]]):
                self.initial_distributions_counts[
                    client_population_metrics_labels[i]][client_population_metrics_values[i]] += 1
            else:
                self.initial_distributions_counts[
                    client_population_metrics_labels[i]][client_population_metrics_values[i]] = 1
        if s.defender_obs_state.aggregated_host_metrics is None:
            raise ValueError("HostMetrics is None")
        aggregated_host_metrics_values, aggregated_host_metrics_labels = \
            s.defender_obs_state.aggregated_host_metrics.get_values()
        for i in range(len(aggregated_host_metrics_values)):
            if (aggregated_host_metrics_values[i] in
                    self.initial_distributions_counts[aggregated_host_metrics_labels[i]]):
                self.initial_distributions_counts[
                    aggregated_host_metrics_labels[i]][aggregated_host_metrics_values[i]] += 1
            else:
                self.initial_distributions_counts[aggregated_host_metrics_labels[i]][
                    aggregated_host_metrics_values[i]] = 1

        for machine in s.defender_obs_state.machines:
            if machine.host_metrics is None:
                raise ValueError("HostMetrics is None")
            m_values, m_labels = machine.host_metrics.get_values()
            for i in range(len(m_values)):
                lbl = f"{m_labels[i]}_{machine.ips[0]}"
                if (m_values[i] in self.initial_distributions_counts[lbl]):
                    self.initial_distributions_counts[lbl][m_values[i]] += 1
                else:
                    self.initial_distributions_counts[lbl][m_values[i]] = 1

    def compute_descriptive_statistics_and_distributions(self) -> None:
        """
        Computes descriptive statistics and empirical probability distributions based on the counters.

        :return: None
        """
        self.num_measurements = 0
        self.conditions = list(self.conditionals_counts.keys())
        self.num_conditions = len(self.conditions)
        for condition in self.conditionals_counts.keys():
            self.means[condition] = {}
            self.stds[condition] = {}
            self.mins[condition] = {}
            self.maxs[condition] = {}
            self.conditionals_probs[condition] = {}
            for metric in self.conditionals_counts[condition].keys():
                self.conditionals_probs[condition][metric] = {}
                observations: List[Any] = []
                total_counts = sum(self.conditionals_counts[condition][metric].values())
                self.num_measurements = self.num_measurements + total_counts
                for value in self.conditionals_counts[condition][metric].keys():
                    tmp = self.conditionals_counts[condition][metric][value] / total_counts
                    self.conditionals_probs[condition][metric][value] = tmp
                    observations = observations + [int(round(float(value)))] * int(
                        round(float(self.conditionals_counts[condition][metric][value])))
                if len(observations) == 0:
                    self.means[condition][metric] = -1
                    self.stds[condition][metric] = -1
                    self.mins[condition][metric] = -1
                    self.maxs[condition][metric] = -1
                else:
                    self.means[condition][metric] = round(float(np.mean(observations)), 2)
                    self.stds[condition][metric] = round(float(np.std(observations)), 2)
                    self.mins[condition][metric] = round(float(np.min(observations)), 2)
                    self.maxs[condition][metric] = round(float(np.max(observations)), 2)
        self.num_metrics = len(self.initial_distributions_counts.keys())
        self.metrics = list(self.initial_distributions_counts.keys())
        for metric in self.initial_distributions_counts.keys():
            self.initial_distributions_probs[metric] = {}
            total_counts = sum(self.initial_distributions_counts[metric].values())
            observations = []
            for value in self.initial_distributions_counts[metric].keys():
                self.initial_distributions_probs[metric][value] = (self.initial_distributions_counts[metric][value] /
                                                                   total_counts)
                observations = (observations + [int(round(float(value)))] * int(round(
                    float(self.initial_distributions_counts[metric][value]))))
            if len(observations) == 0:
                self.initial_means[metric] = -1
                self.initial_stds[metric] = -1
                self.initial_mins[metric] = -1
                self.initial_maxs[metric] = -1
            else:
                self.initial_means[metric] = round(float(np.mean(observations)), 2)
                self.initial_stds[metric] = round(float(np.std(observations)), 2)
                self.initial_mins[metric] = round(float(np.min(observations)), 2)
                self.initial_maxs[metric] = round(float(np.max(observations)), 2)

        for condition1 in list(self.conditionals_counts.keys()):
            self.conditionals_kl_divergences[condition1] = {}
            for condition2 in list(self.conditionals_counts.keys()):
                self.conditionals_kl_divergences[condition1][condition2] = {}
                for metric in self.conditionals_counts[condition1].keys():
                    if (len(list(self.conditionals_probs[condition1][metric].keys())) > 0
                            and len(list(self.conditionals_probs[condition2][metric].keys())) > 0):
                        normalized_p_1 = []
                        normalized_p_2 = []
                        for val in set((list(self.conditionals_probs[condition1][metric].keys()) +
                                        list(self.conditionals_probs[condition2][metric].keys()))):
                            if val in self.conditionals_probs[condition1][metric]:
                                normalized_p_1.append(self.conditionals_probs[condition1][metric][val])
                            else:
                                normalized_p_1.append(0.0)
                            if val in self.conditionals_probs[condition2][metric]:
                                normalized_p_2.append(self.conditionals_probs[condition2][metric][val])
                            else:
                                normalized_p_2.append(0.0)
                        self.conditionals_kl_divergences[condition1][condition2][metric] = \
                            float(round(sum(rel_entr(normalized_p_1, normalized_p_2)), 3))
                        if math.isinf(self.conditionals_kl_divergences[condition1][condition2][metric]):
                            self.conditionals_kl_divergences[condition1][condition2][metric] = "inf"
                    else:
                        self.conditionals_kl_divergences[condition1][condition2][metric] = -1

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"conditionals:{self.conditionals_counts}, initial distributions: {self.initial_distributions_counts}" \
               f"emulation_name: {self.emulation_name}, description: {self.descr}, means: {self.means}, " \
               f"maxs: {self.maxs}, mins: {self.mins}, stds: {self.stds}, " \
               f"conditionals_kl_divergences: {self.conditionals_kl_divergences}, " \
               f"num_measurements: {self.num_measurements}, num_metrics: {self.num_metrics}, metrics: {self.metrics}," \
               f"conditions: {self.conditions}, num_conditions: {self.num_conditions}"

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationStatistics":
        """
        Converts a dict representation of the object to a DTO

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationStatistics(
            emulation_name=d["emulation_name"],
            descr=d["descr"]
        )
        obj.conditionals_counts = d["conditionals_counts"]
        obj.initial_distributions_counts = d["initial_distributions_counts"]
        if "id" in d:
            obj.id = d["id"]
        if "means" in d:
            obj.means = d["means"]
        if "stds" in d:
            obj.stds = d["stds"]
        if "mins" in d:
            obj.mins = d["mins"]
        if "maxs" in d:
            obj.maxs = d["maxs"]
        if "initial_distributions_probs" in d:
            obj.initial_distributions_probs = d["initial_distributions_probs"]
        if "conditionals_probs" in d:
            obj.conditionals_probs = d["conditionals_probs"]
        if "initial_means" in d:
            obj.initial_means = d["initial_means"]
        if "initial_stds" in d:
            obj.initial_stds = d["initial_stds"]
        if "initial_mins" in d:
            obj.initial_mins = d["initial_mins"]
        if "initial_maxs" in d:
            obj.initial_maxs = d["initial_maxs"]
        if "num_metrics" in d:
            obj.num_metrics = d["num_metrics"]
        if "num_measurements" in d:
            obj.num_measurements = d["num_measurements"]
        if "metrics" in d:
            obj.metrics = d["metrics"]
        if "conditions" in d:
            obj.conditions = d["conditions"]
        if "num_conditions" in d:
            obj.num_conditions = d["num_conditions"]
        obj.conditionals_kl_divergences = d["conditionals_kl_divergences"]
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["descr"] = self.descr
        d["id"] = self.id
        d["conditionals_counts"] = self.conditionals_counts
        d["initial_distributions_counts"] = self.initial_distributions_counts
        d["emulation_name"] = self.emulation_name
        d["means"] = self.means
        d["stds"] = self.stds
        d["maxs"] = self.maxs
        d["mins"] = self.mins
        d["conditionals_probs"] = self.conditionals_probs
        d["initial_distributions_probs"] = self.initial_distributions_probs
        d["initial_means"] = self.initial_means
        d["initial_stds"] = self.initial_stds
        d["initial_maxs"] = self.initial_maxs
        d["initial_mins"] = self.initial_mins
        d["conditionals_kl_divergences"] = self.conditionals_kl_divergences
        d["num_metrics"] = self.num_metrics
        d["num_conditions"] = self.num_conditions
        d["num_measurements"] = self.num_measurements
        d["metrics"] = self.metrics
        d["conditions"] = self.conditions
        return d

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationStatistics":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationStatistics.from_dict(json.loads(json_str))

    def get_number_of_samples(self) -> int:
        """
        Counts the number of samples

        :return: the number of samples
        """
        num_samples = 0
        for k, v in self.conditionals_counts.items():
            for k2, v2 in v.items():
                for k3, v3 in v2.items():
                    num_samples += v3
        return num_samples

    def merge(self, second_statistic: "EmulationStatistics") -> None:
        """
        Merges the statistic with another statistic by adding the counts

        :param second_statistic: the statistic to merge with
        :return: None
        """
        for condition in self.conditionals_counts.keys():
            for metric in self.conditionals_counts[condition].keys():
                for value in self.conditionals_counts[condition][metric].keys():
                    if (condition in second_statistic.conditionals_counts
                            and metric in second_statistic.conditionals_counts[condition]
                            and value in second_statistic.conditionals_counts[condition][metric]):
                        self.conditionals_counts[condition][metric][value] = \
                            self.conditionals_counts[condition][metric][value] + \
                            second_statistic.conditionals_counts[condition][metric][value]

        for condition in second_statistic.conditionals_counts.keys():
            if condition not in self.conditionals_counts:
                self.conditionals_counts[condition] = {}
            for metric in second_statistic.conditionals_counts[condition].keys():
                if metric not in self.conditionals_counts[condition]:
                    self.conditionals_counts[condition][metric] = {}
                for value in second_statistic.conditionals_counts[condition][metric].keys():
                    if value not in self.conditionals_counts[condition][metric]:
                        self.conditionals_counts[condition][metric][value] = \
                            second_statistic.conditionals_counts[condition][metric][value]

        for metric in self.initial_distributions_counts.keys():
            for value in self.initial_distributions_counts[metric].keys():
                if (metric in second_statistic.initial_distributions_counts
                        and value in second_statistic.initial_distributions_counts[metric]):
                    self.initial_distributions_counts[metric][value] = \
                        self.initial_distributions_counts[metric][value] + \
                        second_statistic.initial_distributions_counts[metric][value]

        for metric in second_statistic.initial_distributions_counts.keys():
            if metric not in self.initial_distributions_counts:
                self.initial_distributions_counts[metric] = {}
            for value in second_statistic.initial_distributions_counts[metric].keys():
                if value not in self.initial_distributions_counts[metric]:
                    self.initial_distributions_counts[metric][value] = \
                        second_statistic.initial_distributions_counts[metric][value]



from typing import Optional, List
import csle_common.constants.constants as constants
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_outcome import EmulationAttackerActionOutcome


class EmulationAttackerMasscanActions:
    """
    Class containing attacker Massscan actions in the emulation
    """

    @staticmethod
    def MASSCAN_HOST_SCAN(index: int, ips: Optional[List[str]] = None, host_ip: str = "") -> EmulationAttackerAction:
        """
        Action for running a MASSCAN network scan

        :param index: the index of the action
        :param ips: ips of the machines or subnets to apply the action to
        :param host_ip: the host ip
        :return: The created action
        """
        id = EmulationAttackerActionId.MASSCAN_HOST_SCAN
        if ips is None:
            ips = []
        cmd = ["sudo masscan " + constants.MASSCAN.BASE_ARGS + " " + constants.MASSCAN.HOST_ARG + host_ip + " "]
        if index == -1:
            id = EmulationAttackerActionId.MASSCAN_ALL_SCAN

        return EmulationAttackerAction(id=id, name="Masscan port 0-1024", cmds=cmd,
                                       type=EmulationAttackerActionType.RECON,
                                       descr="Masscan port 0-1024",
                                       ips=ips, index=index,
                                       action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING)



from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.metastore.metastore_facade import MetastoreFacade


class SimulationEnvController:
    """
    Class managing operations related to simulation environments
    """

    @staticmethod
    def install_simulation(config: SimulationEnvConfig) -> None:
        """
        Installs the simulation configuration in the metastore

        :param config: the configuration to install
        :return: None
        """
        MetastoreFacade.install_simulation(config=config)

    @staticmethod
    def uninstall_simulation(config: SimulationEnvConfig) -> None:
        """
        Uninstalls a simulation config from the metastore

        :param config: the config to uninstall
        :return: None
        """
        MetastoreFacade.uninstall_simulation(config=config)

    @staticmethod
    def save_simulation_image(img: bytes, simulation: str) -> None:
        """
        Saves the simulation image

        :param image: the image data
        :param simulation: the name of the simulation
        :return: None
        """
        MetastoreFacade.save_simulation_image(img=img, simulation_name=simulation)



from typing import Union, List, Dict, Optional, Tuple, Any
import random
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.training.policy import Policy
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.policy_type import PolicyType


class RandomPolicy(Policy):
    """
    Object representing a static policy
    """

    def __init__(self, player_type: PlayerType, actions: List[Action],
                 stage_policy_tensor: Optional[List[List[float]]]) -> None:
        """
        Initializes the policy

        :param actions: list of actions
        :param player_type: the player type
        :param stage_policy_tensor: the stage policy tensor
        """
        super(RandomPolicy, self).__init__(agent_type=AgentType.RANDOM, player_type=player_type)
        self.actions = actions
        self.stage_policy_tensor = stage_policy_tensor
        self.policy_type = PolicyType.RANDOM

    def action(self, o: Union[List[Union[int, float]], int, float], deterministic: bool = True) -> int:
        """
        Selects the next action

        :param o: the input observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the next action and its probability
        """
        action = random.choice(self.actions)
        return action.id

    def _defender_action(self, o: Union[List[Union[int, float]], int, float]) -> Tuple[Union[int, float], float]:
        """
        Selects the next defender action

        :param o: the input observation
        :return: the next action and its probability
        """
        action = random.choice(self.actions)
        return action.id, self.probability(o=o, a=action.id)

    def probability(self, o: Union[List[Union[int, float]], int, float], a: int) -> float:
        """
        Calculates the probability of taking a given action for a given observation

        :param o: the input observation
        :param a: the action
        :return: p(a|o)
        """
        return 1 / len(self.actions)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "RandomPolicy":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        return RandomPolicy(actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
                            stage_policy_tensor=d["stage_policy_tensor"], player_type=d["player_type"])

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: A dict representation of the function
        """
        d: Dict[str, Any] = {}
        d["agent_type"] = self.agent_type
        d["player_type"] = self.player_type
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["stage_policy_tensor"] = self.stage_policy_tensor
        d["policy_type"] = self.policy_type
        return d

    def stage_policy(self, o: Any) -> Union[List[List[float]], List[float]]:
        """
        Gets the stage policy, i.e a |S|x|A| policy

        :param o: the latest observation
        :return: the |S|x|A| stage policy
        """
        if self.stage_policy_tensor is None:
            raise ValueError("Stage policy is None")
        return self.stage_policy_tensor

    def __str__(self) -> str:
        """
        :return: a string representation of the policy
        """
        return f"agent_type: {self.agent_type}, player_type: {self.player_type}, " \
               f"actions: {list(map(lambda x: str(x), self.actions))}, " \
               f"stage_policy_tensor: {self.stage_policy_tensor}, policy_type: {self.policy_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "RandomPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return RandomPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "RandomPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())



try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer


class ForwardSSHServer(SocketServer.ThreadingTCPServer):
    """
    SSH Server for forwarding local port over a SSH tunnel
    """
    daemon_threads = True
    allow_reuse_address = True



import bcrypt
import csle_common.constants.constants as constants
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.dao.management.management_user import ManagementUser


class ManagementUtil:
    """
    Class with utility functions for management of CSLE
    """

    @staticmethod
    def create_default_management_admin_account() -> None:
        """
        Creates the default management admin account

        :return: None
        """
        management_usernames = list(map(lambda x: x.username, MetastoreFacade.list_management_users()))
        if constants.CSLE_ADMIN.MANAGEMENT_USER not in management_usernames:
            Logger.__call__().get_logger().info(f"Creating management account with administrator privileges, "
                                                f"username: {constants.CSLE_ADMIN.MANAGEMENT_USER}")
            admin = True
            byte_pwd = constants.CSLE_ADMIN.MANAGEMENT_PW.encode('utf-8')
            salt = bcrypt.gensalt()
            pw_hash = bcrypt.hashpw(byte_pwd, salt)
            user = ManagementUser(username=constants.CSLE_ADMIN.MANAGEMENT_USER,
                                  password=pw_hash.decode("utf-8"), admin=admin, salt=salt.decode("utf-8"),
                                  first_name=constants.CSLE_ADMIN.MANAGEMENT_FIRST_NAME,
                                  last_name=constants.CSLE_ADMIN.MANAGEMENT_LAST_NAME,
                                  organization=constants.CSLE_ADMIN.MANAGEMENT_ORGANIZATION,
                                  email=constants.CSLE_ADMIN.MANAGEMENT_EMAIL)
            MetastoreFacade.save_management_user(management_user=user)
        else:
            Logger.__call__().get_logger().info(f"Management account with username: "
                                                f"{constants.CSLE_ADMIN.MANAGEMENT_USER} already exists")

    @staticmethod
    def create_default_management_guest_account() -> None:
        """
        Creates the default management guest account

        :return: None
        """
        management_usernames = list(map(lambda x: x.username, MetastoreFacade.list_management_users()))
        if constants.CSLE_GUEST.MANAGEMENT_USER not in management_usernames:
            Logger.__call__().get_logger().info(f"Creating management account without administrator privileges, "
                                                f"username: {constants.CSLE_GUEST.MANAGEMENT_USER}")
            admin = False
            byte_pwd = constants.CSLE_GUEST.MANAGEMENT_PW.encode('utf-8')
            salt = bcrypt.gensalt()
            pw_hash = bcrypt.hashpw(byte_pwd, salt)
            user = ManagementUser(username=constants.CSLE_GUEST.MANAGEMENT_USER,
                                  password=pw_hash.decode("utf-8"), admin=admin, salt=salt.decode("utf-8"),
                                  first_name=constants.CSLE_GUEST.MANAGEMENT_FIRST_NAME,
                                  last_name=constants.CSLE_GUEST.MANAGEMENT_LAST_NAME,
                                  organization=constants.CSLE_GUEST.MANAGEMENT_ORGANIZATION,
                                  email=constants.CSLE_GUEST.MANAGEMENT_EMAIL)
            MetastoreFacade.save_management_user(management_user=user)
        else:
            Logger.__call__().get_logger().info(f"Management account with username: "
                                                f"{constants.CSLE_GUEST.MANAGEMENT_USER} "
                                                f"already exists")



from typing import Dict, Any, Union
from csle_base.json_serializable import JSONSerializable


class OSSECIDSManagerConfig(JSONSerializable):
    """
    Represents the configuration of the OSSEC IDS managers in a CSLE emulation
    """

    def __init__(self, ossec_ids_manager_log_file: str, ossec_ids_manager_log_dir: str,
                 ossec_ids_manager_max_workers: int,
                 time_step_len_seconds: int = 15, ossec_ids_manager_port: int = 50047,
                 version: str = "0.0.1") -> None:
        """
        Initializes the DTO

        :param time_step_len_seconds: the length of a time-step (period for logging)
        :param version: the version
        :param ossec_ids_manager_port: the GRPC port of the OSSEC IDS manager
        :param ossec_ids_manager_log_file: Log file of the OSSEC IDS manager
        :param ossec_ids_manager_log_dir: Log dir of the OSSEC IDS manager
        :param ossec_ids_manager_max_workers: Max GRPC workers of the OSSEC IDS manager
        """
        self.time_step_len_seconds = time_step_len_seconds
        self.version = version
        self.ossec_ids_manager_port = ossec_ids_manager_port
        self.ossec_ids_manager_log_file = ossec_ids_manager_log_file
        self.ossec_ids_manager_log_dir = ossec_ids_manager_log_dir
        self.ossec_ids_manager_max_workers = ossec_ids_manager_max_workers

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "OSSECIDSManagerConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = OSSECIDSManagerConfig(time_step_len_seconds=d["time_step_len_seconds"],
                                    version=d["version"], ossec_ids_manager_port=d["ossec_ids_manager_port"],
                                    ossec_ids_manager_log_dir=d["ossec_ids_manager_log_dir"],
                                    ossec_ids_manager_log_file=d["ossec_ids_manager_log_file"],
                                    ossec_ids_manager_max_workers=d["ossec_ids_manager_max_workers"])
        return obj

    def to_dict(self) -> Dict[str, Union[str, int]]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, int]] = {}
        d["ossec_ids_manager_port"] = self.ossec_ids_manager_port
        d["time_step_len_seconds"] = self.time_step_len_seconds
        d["version"] = self.version
        d["ossec_ids_manager_log_file"] = self.ossec_ids_manager_log_file
        d["ossec_ids_manager_log_dir"] = self.ossec_ids_manager_log_dir
        d["ossec_ids_manager_max_workers"] = self.ossec_ids_manager_max_workers
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ossec_ids_manager_port: {self.ossec_ids_manager_port}, " \
               f"time_step_len_seconds: {self.time_step_len_seconds}," \
               f" version: {self.version}, ossec_ids_manager_log_file: {self.ossec_ids_manager_log_file}, " \
               f"ossec_ids_manager_log_dir: {self.ossec_ids_manager_log_dir}, " \
               f"ossec_ids_manager_max_workers: {self.ossec_ids_manager_max_workers}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "OSSECIDSManagerConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return OSSECIDSManagerConfig.from_dict(json.loads(json_str))

    def copy(self) -> "OSSECIDSManagerConfig":
        """
        :return: a copy of the DTO
        """
        return OSSECIDSManagerConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "OSSECIDSManagerConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        return config

    @staticmethod
    def schema() -> "OSSECIDSManagerConfig":
        """
        :return: get the schema of the DTO
        """
        return OSSECIDSManagerConfig(ossec_ids_manager_log_file="ossec_ids_manager.log", ossec_ids_manager_log_dir="/",
                                     ossec_ids_manager_max_workers=10)



from typing import Optional, Dict, Any
import csle_common.constants.constants as constants
from csle_common.dao.emulation_observation.common.emulation_port_observation_state import EmulationPortObservationState
from csle_common.dao.emulation_action_result.nmap_port_status import NmapPortStatus
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_action_result.nmap_http_enum import NmapHttpEnum
from csle_common.dao.emulation_action_result.nmap_http_grep import NmapHttpGrep
from csle_common.dao.emulation_action_result.nmap_vulscan import NmapVulscan
from csle_base.json_serializable import JSONSerializable


class NmapPort(JSONSerializable):
    """
    DTO Representing a Port found with a NMAP scan
    """

    def __init__(self, port_id: int, protocol: TransportProtocol, status: NmapPortStatus = NmapPortStatus.DOWN,
                 service_name: str = "none", http_enum: Optional[NmapHttpEnum] = None,
                 http_grep: Optional[NmapHttpGrep] = None, vulscan: Optional[NmapVulscan] = None,
                 service_version: str = "",
                 service_fp: str = ""):
        """
        Initializes the DTO

        :param port_id: the id of the port found
        :param protocol: the protocol of the port
        :param status: the status of the port
        :param service_name: the service running on the port
        :param http_enum: the http enum output
        :param http_grep: the http grep output
        :param vulscan: the vulscan output
        :param service_version: the service version
        :param service_fp: the service fp
        """
        self.port_id = port_id
        self.protocol = protocol
        self.status = status
        self.service_name = service_name
        self.http_enum = http_enum
        self.http_grep = http_grep
        self.vulscan = vulscan
        self.service_version = service_version
        self.service_fp = service_fp

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"port_id:{self.port_id}, protocol:{self.protocol}, status:{self.status}, " \
               f"service_name:{self.service_name}, http_enum:{self.http_enum}, " \
               f"http_grep:{self.http_grep}, vulscan:{self.vulscan}, " \
               f"service_version:{self.service_version}, service_fp:{self.service_fp}"

    def __hash__(self) -> int:
        """
        :return: a hash of the object
        """
        return hash(self.port_id)

    def __eq__(self, other) -> bool:
        """
        Compares equality of the object with another object

        :param other: the object to compare with
        :return: True if equal otherwise False
        """
        return bool(self.port_id == other.port_id)

    def to_obs(self) -> EmulationPortObservationState:
        """
        Converts the object into a PortObservationState

        :return: the created PortObservationState
        """
        open = self.status == NmapPortStatus.UP
        if self.service_name not in constants.SERVICES.service_lookup:
            self.service_name = "unknown"
        hp_enum = ""
        if self.http_enum is not None:
            hp_enum = self.http_enum.output
        hp_grep = ""
        if self.http_grep is not None:
            hp_grep = self.http_grep.output
        vulscan = ""
        if self.vulscan is not None:
            vulscan = self.vulscan.output
        port_obs = EmulationPortObservationState(port=self.port_id, open=open, service=self.service_name,
                                                 protocol=self.protocol, http_enum=hp_enum,
                                                 http_grep=hp_grep, vulscan=vulscan, version=self.service_version,
                                                 fingerprint=self.service_fp)
        return port_obs

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["port_id"] = self.port_id
        d["protocol"] = self.protocol
        d["status"] = self.status
        d["service_name"] = self.service_name
        d["http_enum"] = self.http_enum.to_dict() if self.http_enum is not None else self.http_enum
        d["http_grep"] = self.http_grep.to_dict() if self.http_grep is not None else self.http_grep
        d["vulscan"] = self.vulscan.to_dict() if self.vulscan is not None else self.vulscan
        d["service_version"] = self.service_version
        d["service_fp"] = self.service_fp
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NmapPort":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NmapPort(port_id=d["port_id"], protocol=d["protocol"], status=d["status"],
                       service_name=d["service_name"], http_enum=NmapHttpEnum.from_dict(d["http_enum"]),
                       http_grep=NmapHttpGrep.from_dict(d["http_grep"]), vulscan=NmapVulscan.from_dict(d["vulscan"]),
                       service_version=d["service_version"], service_fp=d["service_fp"])
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "NmapPort":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NmapPort.from_dict(json.loads(json_str))



from typing import Optional, Dict, Any, Union
from csle_common.dao.system_identification.system_model import SystemModel
from csle_common.dao.system_identification.system_identification_config import SystemIdentificationConfig
from csle_common.dao.system_identification.gaussian_mixture_system_model import GaussianMixtureSystemModel
from csle_common.dao.system_identification.empirical_system_model import EmpiricalSystemModel
from csle_common.dao.system_identification.gp_system_model import GPSystemModel
from csle_common.dao.system_identification.mcmc_system_model import MCMCSystemModel
from csle_base.json_serializable import JSONSerializable


class SystemIdentificationJobConfig(JSONSerializable):
    """
    DTO representing a system identification job
    """

    def __init__(self, emulation_env_name: str, emulation_statistics_id: int,
                 progress_percentage: float, pid: int, log_file_path: str,
                 system_identification_config: SystemIdentificationConfig, physical_host_ip: str,
                 descr: str = "", system_model: Optional[SystemModel] = None):
        """
        Initializes the DTO

        :param emulation_env_name: the name of the emulation that the system identification concerns
        :param emulation_statistics_id: the id of the statistics data to train with
        :param progress_percentage: the progress percentage
        :param pid: the pid of the process
        :param log_file_path: path to the log file
        :param descr: a description of the job
        :param system_model: fitted system model
        :param system_identification_config: the config of the system identification algorithm
        :param physical_host_ip: the IP of the physical host where the job is running
        """
        self.emulation_env_name = emulation_env_name
        self.emulation_statistics_id = emulation_statistics_id
        self.progress_percentage = progress_percentage
        self.pid = pid
        self.log_file_path = log_file_path
        self.descr = descr
        self.system_model = system_model
        self.system_identification_config = system_identification_config
        self.id = -1
        self.running = False
        self.physical_host_ip = physical_host_ip

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["emulation_env_name"] = self.emulation_env_name
        d["pid"] = self.pid
        d["progress_percentage"] = self.progress_percentage
        d["emulation_statistics_id"] = self.emulation_statistics_id
        d["descr"] = self.descr
        d["log_file_path"] = self.log_file_path
        if self.system_model is None:
            d["system_model"] = None
        else:
            d["system_model"] = self.system_model.to_dict()
        d["system_identification_config"] = self.system_identification_config.to_dict()
        d["id"] = self.id
        d["running"] = self.running
        d["physical_host_ip"] = self.physical_host_ip
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SystemIdentificationJobConfig":
        """
        Converts a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        # system_model = None
        parse_models = [GaussianMixtureSystemModel.from_dict, EmpiricalSystemModel.from_dict,
                        GPSystemModel.from_dict, MCMCSystemModel.from_dict]
        system_model: Union[None, SystemModel] = None
        for parse_model in parse_models:
            try:
                system_model = parse_model(d['system_model'])
                break
            except Exception:
                pass
        obj = SystemIdentificationJobConfig(
            emulation_env_name=d["emulation_env_name"], pid=d["pid"],
            progress_percentage=d["progress_percentage"], emulation_statistics_id=d["emulation_statistics_id"],
            descr=d["descr"], log_file_path=d["log_file_path"], system_model=system_model,
            system_identification_config=SystemIdentificationConfig.from_dict(d["system_identification_config"]),
            physical_host_ip=d["physical_host_ip"]
        )
        if "id" in d:
            obj.id = d["id"]
        if "running" in d:
            obj.running = d["running"]
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "SystemIdentificationJobConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return SystemIdentificationJobConfig.from_dict(json.loads(json_str))



from csle_common.dao.system_identification.empirical_conditional import EmpiricalConditional
from csle_common.dao.system_identification.empirical_system_model import EmpiricalSystemModel
from csle_common.dao.system_identification.emulation_statistics import EmulationStatistics
from csle_common.dao.system_identification.gaussian_mixture_conditional import GaussianMixtureConditional
from csle_common.dao.system_identification.gaussian_mixture_system_model import GaussianMixtureSystemModel
from csle_common.dao.system_identification.gp_conditional import GPConditional
from csle_common.dao.system_identification.gp_system_model import GPSystemModel
from csle_common.dao.system_identification.mcmc_posterior import MCMCPosterior
from csle_common.dao.system_identification.mcmc_system_model import MCMCSystemModel
from csle_common.dao.system_identification.system_identification_config import SystemIdentificationConfig


class TestSystemIdentificationDaoSuite:
    """
    Test suite for management data access objects (DAOs)
    """

    def test_empirical_conditional(self, example_empirical_conditional: EmpiricalConditional) -> None:
        """
        Tests creation and dict conversion of the EmpiricalConditional DAO

        :param example_empirical_conditional: an example EmpiricalConditional
        :return: None
        """
        assert isinstance(example_empirical_conditional.to_dict(), dict)
        assert isinstance(EmpiricalConditional.from_dict(example_empirical_conditional.to_dict()), EmpiricalConditional)
        assert (EmpiricalConditional.from_dict(example_empirical_conditional.to_dict()).to_dict() ==
                example_empirical_conditional.to_dict())
        assert EmpiricalConditional.from_dict(example_empirical_conditional.to_dict()) == example_empirical_conditional

    def test_empirical_system_model(self, example_empirical_system_model: EmpiricalSystemModel) -> None:
        """
        Tests creation and dict conversion of the EmpiricalSystemModel DAO

        :param example_empirical_system_model: an example EmpiricalSystemModel
        :return: None
        """
        assert isinstance(example_empirical_system_model.to_dict(), dict)
        assert isinstance(EmpiricalSystemModel.from_dict(example_empirical_system_model.to_dict()),
                          EmpiricalSystemModel)
        assert (EmpiricalSystemModel.from_dict(example_empirical_system_model.to_dict()).to_dict() ==
                example_empirical_system_model.to_dict())
        assert EmpiricalSystemModel.from_dict(example_empirical_system_model.to_dict()) == \
               example_empirical_system_model

    def test_emulation_statistics(self, example_emulation_statistics: EmulationStatistics) -> None:
        """
        Tests creation and dict conversion of the EmulationStatistics DAO

        :param example_emulation_statistics: an example EmulationStatistics
        :return: None
        """
        assert isinstance(example_emulation_statistics.to_dict(), dict)
        assert isinstance(EmulationStatistics.from_dict(example_emulation_statistics.to_dict()),
                          EmulationStatistics)
        assert (EmulationStatistics.from_dict(example_emulation_statistics.to_dict()).to_dict() ==
                example_emulation_statistics.to_dict())
        assert EmulationStatistics.from_dict(example_emulation_statistics.to_dict()) == example_emulation_statistics

    def test_gaussian_mixture_conditional(
            self, example_gaussian_mixture_conditional: GaussianMixtureConditional) -> None:
        """
        Tests creation and dict conversion of the GaussianMixtureConditional DAO

        :param example_gaussian_mixture_conditional: an example GaussianMixtureConditional
        :return: None
        """
        assert isinstance(example_gaussian_mixture_conditional.to_dict(), dict)
        assert isinstance(GaussianMixtureConditional.from_dict(example_gaussian_mixture_conditional.to_dict()),
                          GaussianMixtureConditional)
        assert (GaussianMixtureConditional.from_dict(example_gaussian_mixture_conditional.to_dict()).to_dict() ==
                example_gaussian_mixture_conditional.to_dict())
        assert (GaussianMixtureConditional.from_dict(example_gaussian_mixture_conditional.to_dict()) ==
                example_gaussian_mixture_conditional)

    def test_gaussian_mixture_system_model(
            self, example_gaussian_mixture_system_model: GaussianMixtureSystemModel) -> None:
        """
        Tests creation and dict conversion of the GaussianMixtureSystemModel DAO

        :param example_gaussian_mixture_conditional: an example GaussianMixtureConditional
        :return: None
        """
        assert isinstance(example_gaussian_mixture_system_model.to_dict(), dict)
        assert isinstance(GaussianMixtureSystemModel.from_dict(example_gaussian_mixture_system_model.to_dict()),
                          GaussianMixtureSystemModel)
        assert (GaussianMixtureSystemModel.from_dict(example_gaussian_mixture_system_model.to_dict()).to_dict() ==
                example_gaussian_mixture_system_model.to_dict())
        assert (GaussianMixtureSystemModel.from_dict(example_gaussian_mixture_system_model.to_dict()) ==
                example_gaussian_mixture_system_model)

    def test_gp_conditional(self, example_gp_conditional: GPConditional) -> None:
        """
        Tests creation and dict conversion of the GPConditional DAO

        :param example_gp_conditional: an example GPConditional
        :return: None
        """
        assert isinstance(example_gp_conditional.to_dict(), dict)
        assert isinstance(GPConditional.from_dict(example_gp_conditional.to_dict()),
                          GPConditional)
        assert GPConditional.from_dict(example_gp_conditional.to_dict()).to_dict() == example_gp_conditional.to_dict()
        assert GPConditional.from_dict(example_gp_conditional.to_dict()) == example_gp_conditional

    def test_gp_system_model(self, example_gp_system_model: GPSystemModel) -> None:
        """
        Tests creation and dict conversion of the GPSystemModel DAO

        :param example_gp_system_model: an example GPSystemModel
        :return: None
        """
        assert isinstance(example_gp_system_model.to_dict(), dict)
        assert isinstance(GPSystemModel.from_dict(example_gp_system_model.to_dict()), GPSystemModel)
        assert (GPSystemModel.from_dict(example_gp_system_model.to_dict()).to_dict() ==
                example_gp_system_model.to_dict())
        assert (GPSystemModel.from_dict(example_gp_system_model.to_dict()) == example_gp_system_model)

    def test_mcmc_posterior(self, example_mcmc_posterior: MCMCPosterior) -> None:
        """
        Tests creation and dict conversion of the MCMCPosterior DAO

        :param example_mcmc_posterior: an example MCMCPosterior
        :return: None
        """
        assert isinstance(example_mcmc_posterior.to_dict(), dict)
        assert isinstance(MCMCPosterior.from_dict(example_mcmc_posterior.to_dict()), MCMCPosterior)
        assert MCMCPosterior.from_dict(example_mcmc_posterior.to_dict()).to_dict() == example_mcmc_posterior.to_dict()
        assert MCMCPosterior.from_dict(example_mcmc_posterior.to_dict()) == example_mcmc_posterior

    def test_mcmc_system_model(self, example_mcmc_system_model: MCMCSystemModel) -> None:
        """
        Tests creation and dict conversion of the MCMCSystemModel DAO

        :param example_mcmc_system_model: an example MCMCSystemModel
        :return: None
        """
        assert isinstance(example_mcmc_system_model.to_dict(), dict)
        assert isinstance(MCMCSystemModel.from_dict(example_mcmc_system_model.to_dict()),
                          MCMCSystemModel)
        assert (MCMCSystemModel.from_dict(example_mcmc_system_model.to_dict()).to_dict() ==
                example_mcmc_system_model.to_dict())
        assert MCMCSystemModel.from_dict(example_mcmc_system_model.to_dict()) == example_mcmc_system_model

    def test_system_identification_config(
            self, example_system_identification_config: SystemIdentificationConfig) -> None:
        """
        Tests creation and dict conversion of the SystemIdentificationConfig DAO

        :param example_system_identification_config: an example SystemIdentificationConfig
        :return: None
        """
        assert isinstance(example_system_identification_config.to_dict(), dict)
        assert isinstance(SystemIdentificationConfig.from_dict(example_system_identification_config.to_dict()),
                          SystemIdentificationConfig)
        assert (SystemIdentificationConfig.from_dict(example_system_identification_config.to_dict()).to_dict() ==
                example_system_identification_config.to_dict())
        assert (SystemIdentificationConfig.from_dict(example_system_identification_config.to_dict()) ==
                example_system_identification_config)



from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_action.defender.emulation_defender_action_type import EmulationDefenderActionType
from csle_common.dao.emulation_action.defender.emulation_defender_action_id import EmulationDefenderActionId
from csle_common.dao.emulation_action.defender.emulation_defender_action_outcome import EmulationDefenderActionOutcome
from typing import List


class EmulationDefenderStoppingActions:
    """
    Class implementing stopping actions for the defender
    """

    @staticmethod
    def STOP(index: int) -> EmulationDefenderAction:
        """
        Reports a detected intrusion and stops

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationDefenderActionId.STOP
        cmd: List[str] = []
        alt_cmd: List[str] = []
        return EmulationDefenderAction(id=id, name="Report Intrusion", cmds=cmd,
                                       type=EmulationDefenderActionType.STOP,
                                       descr="Reports an ongoing intrusion to the infrastructure and stops",
                                       index=index,
                                       ips=[], action_outcome=EmulationDefenderActionOutcome.GAME_END, alt_cmds=alt_cmd)

    @staticmethod
    def CONTINUE(index: int) -> EmulationDefenderAction:
        """
        A "continue" action, the defender chooses to not make any action

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationDefenderActionId.CONTINUE
        cmd: List[str] = []
        alt_cmd: List[str] = []
        return EmulationDefenderAction(
            id=id, name="Continue", cmds=cmd, type=EmulationDefenderActionType.CONTINUE,
            descr="A 'continue' action, the defender chooses to not make any action", index=index, ips=[],
            action_outcome=EmulationDefenderActionOutcome.CONTINUE, alt_cmds=alt_cmd)

    @staticmethod
    def RESET_USERS(index: int) -> EmulationDefenderAction:
        """
        A non terminal stop action, the defender resets all user accounts, which means that password vulnerabilities
        are mitigated.

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationDefenderActionId.RESET_USERS
        cmd: List[str] = []
        alt_cmd: List[str] = []
        return EmulationDefenderAction(
            id=id, name="Reset Users", cmds=cmd, type=EmulationDefenderActionType.ADD_DEFENSIVE_MECHANISM,
            descr="A non terminal stop action the defender resets all user accounts, which means that "
                  "password vulnerabilities are mitigated", index=index, ips=[],
            action_outcome=EmulationDefenderActionOutcome.ADD_DEFENSIVE_MECHANISM, alt_cmds=alt_cmd)

    @staticmethod
    def ENABLE_DPI(index: int) -> EmulationDefenderAction:
        """
        A non terminal stop action, the defender enables DPI by starting the IDS

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationDefenderActionId.ENABLE_DPI
        cmd: List[str] = []
        alt_cmd: List[str] = []
        return EmulationDefenderAction(id=id, name="Enable DPI", cmds=cmd,
                                       type=EmulationDefenderActionType.ADD_DEFENSIVE_MECHANISM,
                                       descr="A non terminal stop action, the defender enables DPI by starting the IDS",
                                       index=index,
                                       ips=[], action_outcome=EmulationDefenderActionOutcome.ADD_DEFENSIVE_MECHANISM,
                                       alt_cmds=alt_cmd)

    @staticmethod
    def BLACKLIST_IPS(index: int) -> EmulationDefenderAction:
        """
        A non terminal stop action, the defender blacklists all IPs that generate IDS alerts that exceed a threshold,
        the blacklist is implemented using the firewall.

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationDefenderActionId.BLACKLIST_IPS
        cmd: List[str] = []
        alt_cmd: List[str] = []
        return EmulationDefenderAction(
            id=id, name="Blacklist IPs", cmds=cmd,
            type=EmulationDefenderActionType.ADD_DEFENSIVE_MECHANISM,
            descr="A non terminal stop action, the defender blacklists all IPs that generate "
                  "IDS alerts that exceed a threshold, the blacklist is implemented using the firewall.",
            index=index, ips=[], action_outcome=EmulationDefenderActionOutcome.ADD_DEFENSIVE_MECHANISM,
            alt_cmds=alt_cmd)



from typing import Dict, Any, List
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.emulation_config.node_resources_config import NodeResourcesConfig
from csle_common.dao.emulation_config.node_firewall_config import NodeFirewallConfig
from csle_base.json_serializable import JSONSerializable


class ElkConfig(JSONSerializable):
    """
    Represents the configuration of an ELK node in CSLE
    """

    def __init__(self, container: NodeContainerConfig, resources: NodeResourcesConfig,
                 firewall_config: NodeFirewallConfig,
                 elk_manager_log_file: str, elk_manager_log_dir: str, elk_manager_max_workers: int,
                 elastic_port: int = 9200, kibana_port: int = 5601, logstash_port: int = 5044,
                 time_step_len_seconds: int = 15, elk_manager_port: int = 50045, version: str = "0.0.1") -> None:
        """
        Initializes the DTO

        :param container: the container for the ELK stack
        :param network: the network
        :param elastic_port: the port that the Elastic db server is listening to
        :param kibana_port: the port that the kibana web server listens to
        :param logstash_port: the port that the logstash beat interface listens to
        :param elk_manager_port: the default port for the gRPC manager
        :param time_step_len_seconds: the length of a time-step (period for logging)
        :param firewall_config: the firewall configuration
        :param container: the container
        :param version: the version
        :param elk_manager_log_file: the log file of the elk manager
        :param elk_manager_log_dir: the log dir of the elk manager
        :param elk_manager_max_workers: the maximum number of GRPC workers for the elk maanger
        """
        self.elastic_port = elastic_port
        self.kibana_port = kibana_port
        self.logstash_port = logstash_port
        self.elk_manager_port = elk_manager_port
        self.time_step_len_seconds = time_step_len_seconds
        self.version = version
        self.container = container
        self.resources = resources
        self.firewall_config = firewall_config
        self.elk_manager_log_file = elk_manager_log_file
        self.elk_manager_log_dir = elk_manager_log_dir
        self.elk_manager_max_workers = elk_manager_max_workers

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ElkConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ElkConfig(
            container=NodeContainerConfig.from_dict(d["container"]),
            resources=NodeResourcesConfig.from_dict(d["resources"]),
            elastic_port=d["elastic_port"], kibana_port=d["kibana_port"], logstash_port=d["logstash_port"],
            time_step_len_seconds=d["time_step_len_seconds"],
            elk_manager_port=d["elk_manager_port"],
            version=d["version"],
            firewall_config=NodeFirewallConfig.from_dict(d["firewall_config"]),
            elk_manager_max_workers=d["elk_manager_max_workers"],
            elk_manager_log_dir=d["elk_manager_log_dir"],
            elk_manager_log_file=d["elk_manager_log_file"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["container"] = self.container.to_dict()
        d["resources"] = self.resources.to_dict()
        d["elastic_port"] = self.elastic_port
        d["kibana_port"] = self.kibana_port
        d["logstash_port"] = self.logstash_port
        d["elk_manager_port"] = self.elk_manager_port
        d["time_step_len_seconds"] = self.time_step_len_seconds
        d["version"] = self.version
        d["firewall_config"] = self.firewall_config.to_dict()
        d["elk_manager_log_file"] = self.elk_manager_log_file
        d["elk_manager_log_dir"] = self.elk_manager_log_dir
        d["elk_manager_max_workers"] = self.elk_manager_max_workers
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"container: {self.container}, " \
               f"elastic port:{self.elastic_port}, version: {self.version}, resources: {self.resources}, " \
               f"kibana port: {self.kibana_port}, logstash_port: {self.logstash_port} " \
               f"elk_manager_port:{self.elk_manager_port}, time_step_len_seconds: {self.time_step_len_seconds}, " \
               f"firewall_config: {self.firewall_config}, " \
               f"elk_manager_log_file: {self.elk_manager_log_file}, " \
               f"elk_manager_log_dir: {self.elk_manager_log_dir}, " \
               f"elk_manager_max_workers: {self.elk_manager_max_workers}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ElkConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ElkConfig.from_dict(json.loads(json_str))

    def copy(self) -> "ElkConfig":
        """
        :return: a copy of the DTO
        """
        return ElkConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int, physical_servers: List[str]) -> "ElkConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :param physical_servers: the physical servers where the execution is deployedn
        :return: the new config
        """
        config = self.copy()
        config.container = config.container.create_execution_config(ip_first_octet=ip_first_octet,
                                                                    physical_servers=physical_servers)
        config.resources = config.resources.create_execution_config(ip_first_octet=ip_first_octet)
        config.firewall_config = config.firewall_config.create_execution_config(ip_first_octet=ip_first_octet)
        return config

    @staticmethod
    def schema() -> "ElkConfig":
        """
        :return: get the schema of the DTO
        """
        return ElkConfig(container=NodeContainerConfig.schema(), resources=NodeResourcesConfig.schema(),
                         firewall_config=NodeFirewallConfig.schema(), elk_manager_log_file="elk_manager.log",
                         elk_manager_log_dir="/", elk_manager_max_workers=10)



from typing import Union, Tuple
import torch
import torch.nn as nn
from torch.distributions.categorical import Categorical
import numpy as np


class PPONetwork(nn.Module):
    """
    Class for instantiating a neural network for PPO training
    """

    def __init__(self, input_dim: int, output_dim_critic: int, output_dim_action: int,
                 num_hidden_layers: int, hidden_layer_dim: int, std_critic: float = 1.0,
                 std_action: float = 0.01) -> None:
        """
        Initializes the neural network

        :param input_dim: the dimension of the input
        :param output_dim_critic: the dimension of the critic output (generally 1)
        :param output_dim_action: the dimension of the actor output (action space dimension)
        :param num_hidden_layers: the number of hidden layers
        :param hidden_layer_dim: the dimension of a hidden layer
        :param std_critic: the standard deviation of the critic for sampling
        :param std_action: the standard deviation of the actor for sampling
        """
        super(PPONetwork, self).__init__()
        self.input_dim = input_dim
        self.output_dim_critic = output_dim_critic
        self.output_dim_action = output_dim_action
        self.std_critic = std_critic
        self.std_action = std_action
        self.critic = nn.Sequential()
        self.actor = nn.Sequential()
        self.aux_critic = nn.Sequential()
        self.num_hidden_layers = num_hidden_layers
        self.hidden_layer_dim = hidden_layer_dim
        input_dim = self.input_dim
        for layer in range(num_hidden_layers):
            self.critic.add_module(name=f'Layer {layer}', module=self.layer_init(nn.Linear(input_dim,
                                                                                           hidden_layer_dim)))
            self.critic.add_module(name='activation', module=nn.Tanh())
            self.aux_critic.add_module(name=f'Layer {layer}', module=self.layer_init(nn.Linear(input_dim,
                                                                                               hidden_layer_dim)))
            self.aux_critic.add_module(name='activation', module=nn.Tanh())
            self.actor.add_module(name=f'Layer {layer}', module=self.layer_init(nn.Linear(input_dim,
                                                                                          hidden_layer_dim)))
            self.actor.add_module(name='activation', module=nn.Tanh())
            input_dim = hidden_layer_dim
        self.critic.add_module(name='Classifier',
                               module=self.layer_init(nn.Linear(hidden_layer_dim, self.output_dim_critic),
                                                      std=self.std_critic))
        self.aux_critic.add_module(name='Classifier',
                                   module=self.layer_init(nn.Linear(hidden_layer_dim, self.output_dim_critic),
                                                          std=self.std_critic))
        self.actor.add_module(name='Classifier',
                              module=self.layer_init(nn.Linear(hidden_layer_dim, self.output_dim_action),
                                                     std=self.std_action))

    def layer_init(self, layer: nn.Linear, std: float = np.sqrt(2), bias_const: float = 0.0) -> nn.Linear:
        """
        Initializes a layer in the neural network

        :param layer: the layer object
        :param std: the standard deviation
        :param bias_const: the bias constant
        :return: the initialized layer
        """
        torch.nn.init.orthogonal_(layer.weight, std)
        torch.nn.init.constant_(layer.bias, bias_const)
        return layer

    def get_value(self, x: torch.Tensor) -> torch.Tensor:
        """
        Computes the value function V(x)

        :param x: the input observation
        :return: The value
        """
        value: torch.Tensor = self.critic(x)
        return value

    def get_action_and_value(self, x: torch.Tensor, action: Union[torch.Tensor, None] = None) \
            -> Tuple[torch.Tensor, float, torch.Tensor, torch.Tensor]:
        """
        Gets the action and the value prediction of the network for a given input tensor x

        :param x: the input tensor
        :param action: (optional) the action; if not specified the action is sampled
        :return: the action, log p(action), the entropy of the action, V(x)
        """
        logits = self.actor(x)
        probs = Categorical(logits=logits)
        if action is None:
            action = probs.sample()
        return action, probs.log_prob(action), probs.entropy(), self.critic(x)

    def get_pi(self, x: torch.Tensor) -> torch.distributions.Categorical:
        """
        Utility function for PPG

        :param x: the input vector
        :return: the output action distribution
        """
        return Categorical(logits=self.actor(x))

    def get_pi_value_and_aux_value(self, x: torch.Tensor) \
            -> Tuple[torch.distributions.Categorical, torch.Tensor, torch.Tensor]:
        """
        Utility function for PPG

        :param x: the input vector
        :return: output distribution, critic value, and auxiliary critic value
        """
        return Categorical(logits=self.actor(x)), self.critic(x.detach()), self.aux_critic(x)

    def save(self, path: str) -> None:
        """
        Saves the model to disk

        :param path: the path on disk to save the model
        :return: None
        """
        state_dict = self.state_dict()
        state_dict["input_dim"] = self.input_dim
        state_dict["output_dim_critic"] = self.output_dim_critic
        state_dict["output_dim_action"] = self.output_dim_action
        state_dict["std_critic"] = self.std_critic
        state_dict["std_action"] = self.std_action
        state_dict["num_hidden_layers"] = self.num_hidden_layers
        state_dict["hidden_layer_dim"] = self.hidden_layer_dim
        torch.save(state_dict, path)

    @staticmethod
    def load(path: str) -> "PPONetwork":
        """
        Loads the model from a given path

        :param path: the path to load the model from
        :return: None
        """
        state_dict = torch.load(path)
        model = PPONetwork(input_dim=state_dict["input_dim"], output_dim_action=state_dict["output_dim_action"],
                           output_dim_critic=state_dict["output_dim_critic"],
                           num_hidden_layers=state_dict["num_hidden_layers"],
                           hidden_layer_dim=state_dict["hidden_layer_dim"], std_critic=state_dict["std_critic"],
                           std_action=state_dict["std_action"])
        del state_dict["input_dim"]
        del state_dict["output_dim_critic"]
        del state_dict["output_dim_action"]
        del state_dict["std_critic"]
        del state_dict["std_action"]
        del state_dict["num_hidden_layers"]
        del state_dict["hidden_layer_dim"]
        model.load_state_dict(state_dict)
        model.eval()
        return model



import logging
from unittest.mock import patch, MagicMock
import csle_common.constants.constants as constants
from csle_common.controllers.kafka_controller import KafkaController
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
import csle_collector.kafka_manager.kafka_manager_pb2_grpc
import csle_collector.kafka_manager.kafka_manager_pb2
import csle_collector.kafka_manager.query_kafka_server
import csle_collector.kafka_manager.kafka_manager_util


class TestKafkaControllerSuite:
    """
    Test suite for kafka controller
    """

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_start_stop_kafka_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test case for starting or stopping the kafka manager

        :param mock_disconnect_admin: Mocked disconnect_admin method
        :param mock_execute_ssh_cmd: Mocked execute_ssh_cmd method
        :param mock_connect_admin: Mocked connect_admin
        :param mock_sleep: Mocked sleep method
        :return: None
        """
        mock_connect_admin.return_value = None
        mock_execute_ssh_cmd.return_value = ("output", "error", 0)
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        kafka_config = MagicMock()
        emulation_env_config.kafka_config = kafka_config
        emulation_env_config.kafka_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.kafka_config.get_connection.return_value = MagicMock()
        KafkaController.start_kafka_manager(emulation_env_config)
        mock_connect_admin.assert_called_once_with(emulation_env_config=emulation_env_config, ip="172.17.0.1",
                                                   create_producer=False)
        mock_execute_ssh_cmd.assert_called()
        KafkaController.stop_kafka_manager(emulation_env_config)
        assert 2 == mock_connect_admin.call_count
        assert 3 == mock_execute_ssh_cmd.call_count

    @patch("csle_collector.kafka_manager.query_kafka_server.create_topic")
    @patch("csle_collector.kafka_manager.query_kafka_server.start_kafka")
    @patch("csle_collector.kafka_manager.kafka_manager_pb2_grpc.KafkaManagerStub")
    @patch("csle_common.controllers.kafka_controller.KafkaController.start_kafka_manager")
    @patch("csle_common.controllers.kafka_controller.KafkaController.get_kafka_status_by_port_and_ip")
    @patch("time.sleep", return_value=None)
    def test_create_topics(self, mock_time_sleep, mock_get_kafka_status_by_port_and_ip, mock_start_kafka_manager,
                           mock_KafkaManagerStub, mock_start_kafka, mock_create_topic) -> None:
        """
        Tests the method that sends a request to the KafkaManager to create topics according to the given configuration

        :param mock_time_sleep: mock time_sleep method
        :param mock_get_kafka_status_by_port_and_ip: mock get_kafka_status_by_port_and_ip method
        :param mock_start_kafka_manager: mock start_kafka_manager method
        :param mock_KafkaManagerStub: mock KafkaManagerStub
        :param mock_start_kafka: mock start_kafka
        :param mock_create_topic: mock create_topic
        :return: None
        """
        mock_stub = MagicMock()
        mock_KafkaManagerStub.return_value = mock_stub
        mock_logger = MagicMock(spec=logging.Logger)
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        kafka_config = MagicMock()
        kafka_container_mock = MagicMock()
        kafka_container_mock.docker_gw_bridge_ip = "127.0.0.1"
        kafka_container_mock.get_ips.return_value = ["127.0.0.1"]
        kafka_container_mock.get_full_name.return_value = "container_name"
        kafka_config.container = kafka_container_mock
        kafka_config.kafka_manager_port = 9092
        emulation_env_config.kafka_config = kafka_config
        topic1 = MagicMock()
        topic1.name = "topic1"
        topic1.num_partitions = 1
        topic1.num_replicas = 1
        topic1.retention_time_hours = 24
        topic2 = MagicMock()
        topic2.name = "topic2"
        topic2.num_partitions = 2
        topic2.num_replicas = 2
        topic2.retention_time_hours = 48
        emulation_env_config.kafka_config.topics = [topic1, topic2]
        mock_get_kafka_status_by_port_and_ip.return_value.running = False
        KafkaController.create_topics(emulation_env_config=emulation_env_config, logger=mock_logger)
        mock_logger.info.assert_any_call(f"Creating kafka topics on container: "
                                         f"{emulation_env_config.kafka_config.container.docker_gw_bridge_ip} "
                                         f"({kafka_container_mock.get_ips()[0]}, "
                                         f"{kafka_container_mock.get_full_name()})")
        mock_logger.info.assert_any_call("Kafka server is not running, starting it.")
        mock_logger.info.assert_any_call("Creating topic: topic1")
        mock_logger.info.assert_any_call("Creating topic: topic2")
        mock_start_kafka_manager.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_get_kafka_status_by_port_and_ip.assert_called_once_with(ip="127.0.0.1", port=9092)
        mock_start_kafka.assert_called_once_with(mock_stub)
        mock_create_topic.assert_any_call(mock_stub, name="topic1", partitions=1, replicas=1, retention_time_hours=24)
        mock_create_topic.assert_any_call(mock_stub, name="topic2", partitions=2, replicas=2, retention_time_hours=48)
        mock_time_sleep.assert_called_once_with(20)

    @patch("csle_common.controllers.kafka_controller.KafkaController.start_kafka_manager")
    @patch("csle_common.controllers.kafka_controller.KafkaController.get_kafka_status_by_port_and_ip")
    def test_get_kafka_status(self, mock_get_kafka_status_by_port_and_ip, mock_start_kafka_manager) -> None:
        """
        Test case for querying the KafkaManager about the status of the Kafka server

        :param mock_get_kafka_status_by_port_and_ip: mocked get_kafka_status_by_port_and_ip method
        :param mock_start_kafka_manager: mocked start_kafka_manager method
        :return: None
        """
        mock_start_kafka_manager.return_value = None
        expected_kafka_dto = MagicMock(spec=csle_collector.kafka_manager.kafka_manager_pb2.KafkaDTO)
        mock_get_kafka_status_by_port_and_ip.return_value = expected_kafka_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        kafka_config = MagicMock()
        emulation_env_config.kafka_config = kafka_config
        emulation_env_config.kafka_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.kafka_config.kafka_manager_port = 5601
        result = KafkaController.get_kafka_status(emulation_env_config)
        mock_start_kafka_manager.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_get_kafka_status_by_port_and_ip.assert_called_once_with(ip="172.17.0.1", port=5601)
        assert result == expected_kafka_dto

    @patch("csle_collector.kafka_manager.query_kafka_server.get_kafka_status")
    @patch("csle_collector.kafka_manager.kafka_manager_pb2_grpc.KafkaManagerStub")
    @patch("grpc.insecure_channel")
    def test_get_kafka_status_by_port_and_ip(self, mock_insecure_channel, mock_KafkaManagerStub,
                                             mock_get_kafka_status) -> None:
        """
        Test case for querying the KafkaManager about the status of the kafka stack

        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_KafkaManagerStub: mocked KafkaManagerStub method
        :param mock_get_kafka_status: mocked get_kafka_status method
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_KafkaManagerStub.return_value = mock_stub
        expected_kafka_dto = MagicMock(spec=csle_collector.kafka_manager.kafka_manager_pb2.KafkaDTO)
        mock_get_kafka_status.return_value = expected_kafka_dto
        ip = "172.17.0.1"
        port = 5601
        timeout = 5
        result = KafkaController.get_kafka_status_by_port_and_ip(ip=ip, port=port, timeout=timeout)
        mock_insecure_channel.assert_called_once_with(f"{ip}:{port}", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_KafkaManagerStub.assert_called_once_with(mock_channel)
        mock_get_kafka_status.assert_called_once_with(mock_stub, timeout=timeout)
        assert result == expected_kafka_dto

    @patch("csle_common.controllers.kafka_controller.KafkaController.start_kafka_manager")
    @patch("csle_collector.kafka_manager.query_kafka_server.stop_kafka")
    @patch("csle_collector.kafka_manager.kafka_manager_pb2_grpc.KafkaManagerStub")
    @patch("grpc.insecure_channel")
    def test_stop_kafka_server(self, mock_insecure_channel, mock_KafkaManagerStub, mock_stop_kafka,
                               mock_start_kafka_manager) -> None:
        """
        Test the method for requesting the KafkaManager to stop the Kafka server

        :param mock_insecure_channel: mock insecure_channel
        :param mock_KafkaManagerStub: mock KafkaManagerStub
        :param mock_stop_kafka: mock stop_kafka
        :param mock_start_kafka_manager: mock start_kafka_manager
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_KafkaManagerStub.return_value = mock_stub
        expected_kafka_dto = MagicMock(spec=csle_collector.kafka_manager.kafka_manager_pb2.KafkaDTO)
        mock_stop_kafka.return_value = expected_kafka_dto
        mock_logger = MagicMock(spec=logging.Logger)
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        kafka_config = MagicMock()
        kafka_container_mock = MagicMock()
        kafka_container_mock.docker_gw_bridge_ip = "127.0.0.1"
        kafka_container_mock.get_ips.return_value = ["127.0.0.1"]
        kafka_container_mock.get_full_name.return_value = "container_name"
        kafka_config.container = kafka_container_mock
        kafka_config.kafka_manager_port = 9092
        emulation_env_config.kafka_config = kafka_config
        result = KafkaController.stop_kafka_server(emulation_env_config=emulation_env_config, logger=mock_logger)
        mock_logger.info.assert_any_call(f"Stopping Kafka server on container: "
                                         f"{kafka_container_mock.docker_gw_bridge_ip} "
                                         f"({kafka_container_mock.get_ips()[0]}, "
                                         f"{kafka_container_mock.get_full_name()})")
        mock_start_kafka_manager.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_insecure_channel.assert_called_once_with(
            f"{emulation_env_config.kafka_config.container.docker_gw_bridge_ip}:"
            f"{emulation_env_config.kafka_config.kafka_manager_port}", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_KafkaManagerStub.assert_called_once_with(mock_channel)
        mock_stop_kafka.assert_called_once_with(mock_stub)
        assert result == expected_kafka_dto

    @patch("csle_common.controllers.kafka_controller.KafkaController.start_kafka_manager")
    @patch("csle_collector.kafka_manager.query_kafka_server.start_kafka")
    @patch("csle_collector.kafka_manager.kafka_manager_pb2_grpc.KafkaManagerStub")
    @patch("grpc.insecure_channel")
    def test_start_kafka_server(self, mock_insecure_channel, mock_KafkaManagerStub, mock_start_kafka,
                                mock_start_kafka_manager) -> None:
        """
        Test the method for requesting the KafkaManager to start the Kafka server

        :param mock_insecure_channel: mock insecure_channel
        :param mock_KafkaManagerStub: mock KafkaManagerStub
        :param mock_start_kafka: mock start_kafka
        :param mock_start_kafka_manager: mock start_kafka_manager
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_KafkaManagerStub.return_value = mock_stub
        expected_kafka_dto = MagicMock(spec=csle_collector.kafka_manager.kafka_manager_pb2.KafkaDTO)
        mock_start_kafka.return_value = expected_kafka_dto
        mock_logger = MagicMock(spec=logging.Logger)
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        kafka_config = MagicMock()
        kafka_container_mock = MagicMock()
        kafka_container_mock.docker_gw_bridge_ip = "127.0.0.1"
        kafka_container_mock.get_ips.return_value = ["127.0.0.1"]
        kafka_container_mock.get_full_name.return_value = "container_name"
        kafka_config.container = kafka_container_mock
        kafka_config.kafka_manager_port = 9092
        emulation_env_config.kafka_config = kafka_config
        result = KafkaController.start_kafka_server(emulation_env_config=emulation_env_config, logger=mock_logger)
        mock_logger.info.assert_any_call(f"Starting Kafka server on container: "
                                         f"{kafka_config.container.docker_gw_bridge_ip} ("
                                         f"{kafka_container_mock.get_ips()[0]}, "
                                         f"{kafka_container_mock.get_full_name()})")
        mock_start_kafka_manager.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_insecure_channel.assert_called_once_with(
            f"{emulation_env_config.kafka_config.container.docker_gw_bridge_ip}:"
            f"{emulation_env_config.kafka_config.kafka_manager_port}", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_KafkaManagerStub.assert_called_once_with(mock_channel)
        mock_start_kafka.assert_called_once_with(mock_stub)
        assert result == expected_kafka_dto

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_collector.constants.constants.KAFKA.KAFKA_CONFIG_FILE")
    def test_configure_broker_ips(self, mock_kafka_config_file, mock_connect_admin) -> None:
        """
        Test the method for configuring the broker IPs on the Kafka container

        :param mock_kafka_config_file: mock kafka_config_file
        :param mock_connect_admin: mock_connect_admin
        :return: None
        """
        mock_logger = MagicMock(spec=logging.Logger)
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        kafka_config = MagicMock()
        kafka_container_mock = MagicMock()
        kafka_container_mock.docker_gw_bridge_ip = "127.0.0.1"
        kafka_container_mock.get_ips.return_value = ["127.0.0.1"]
        kafka_container_mock.get_full_name.return_value = "container_name"
        kafka_config.container = kafka_container_mock
        kafka_config.kafka_manager_port = 9092
        emulation_env_config.kafka_config = kafka_config
        mock_sftp_client = MagicMock()
        mock_file = MagicMock()
        emulation_env_config.get_connection.return_value.open_sftp.return_value = (mock_sftp_client)
        mock_file.read.return_value = b"internal_ip_placeholder external_ip_placeholder"
        mock_sftp_client.open.side_effect = [mock_file, mock_file]  # for read and write
        KafkaController.configure_broker_ips(emulation_env_config=emulation_env_config, logger=mock_logger)
        mock_logger.info.assert_any_call(f"Configuring broker IPs on container: "
                                         f"{kafka_container_mock.docker_gw_bridge_ip} "
                                         f"({kafka_container_mock.get_ips()[0]}, "
                                         f"{kafka_container_mock.get_full_name()})")
        mock_connect_admin.assert_called_once_with(emulation_env_config=emulation_env_config,
                                                   ip=kafka_container_mock.docker_gw_bridge_ip)
        emulation_env_config.get_connection.assert_called_once_with(ip=kafka_container_mock.docker_gw_bridge_ip)
        mock_sftp_client.open.assert_any_call(mock_kafka_config_file, mode="r")
        mock_sftp_client.open.assert_any_call(mock_kafka_config_file, mode="w")
        mock_file.flush.assert_called_once()
        mock_file.close.assert_called()

    def test_get_kafka_managers_ips(self) -> None:
        """
        Test case for extracting the IPS of the kafka managers in a given emulation

        :return: None
        """
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        kafka_config = MagicMock()
        emulation_env_config.kafka_config = kafka_config
        emulation_env_config.kafka_config.container.docker_gw_bridge_ip = "172.17.0.1"
        result = KafkaController.get_kafka_managers_ips(emulation_env_config)
        assert result == ["172.17.0.1"]

    def test_get_kafka_managers_ports(self) -> None:
        """
        Test case for extracting the ports of the kafka managers in a given emulation

        :return: None
        """
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        kafka_config = MagicMock()
        emulation_env_config.kafka_config = kafka_config
        emulation_env_config.kafka_config.kafka_manager_port = 5601
        result = KafkaController.get_kafka_managers_ports(emulation_env_config)
        assert result == [5601]

    @patch("csle_common.controllers.kafka_controller.KafkaController.get_kafka_managers_ips")
    @patch("csle_common.controllers.kafka_controller.KafkaController.get_kafka_managers_ports")
    @patch("csle_common.controllers.kafka_controller.KafkaController.get_kafka_status_by_port_and_ip")
    @patch("csle_collector.kafka_manager.kafka_manager_util.KafkaManagerUtil.kafka_dto_empty")
    def test_get_kafka_managers_infos(self, mock_kafka_dto_empty, mock_get_kafka_status_by_port_and_ip,
                                      mock_get_kafka_managers_ports, mock_get_kafka_managers_ips) -> None:
        """
        Test case for extracting the infomation of the kafka managers for a given emulation

        :param mock_kafka_dto_empty: mocked kafka_dto_empty method
        :param mock_get_kafka_status_by_port_and_ip: mocked get_kafka_status_by_port_and_ip method
        :param mock_get_kafka_managers_ports: mocked get_kafka_managers_ports
        :param mock_get_kafka_managers_ips: mocked get_kafka_managers_ips
        :return: None
        """
        mock_get_kafka_managers_ips.return_value = ["172.17.0.1"]
        mock_get_kafka_managers_ports.return_value = [5601]
        mock_kafka_status = MagicMock()
        mock_get_kafka_status_by_port_and_ip.return_value = mock_kafka_status
        mock_kafka_dto_empty.return_value = MagicMock()
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        kafka_config = MagicMock()
        emulation_env_config.kafka_config = kafka_config
        emulation_env_config.kafka_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.kafka_config.kafka_manager_port = 5601
        emulation_env_config.execution_id = 10
        emulation_env_config.name = "test_emulation"
        logger = MagicMock(spec=logging.Logger)
        active_ips = ["172.17.0.1"]
        physical_host_ip = "192.168.0.1"
        result = KafkaController.get_kafka_managers_info(emulation_env_config, active_ips, logger, physical_host_ip)
        mock_get_kafka_managers_ips.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_get_kafka_managers_ports.assert_called_once_with(emulation_env_config=emulation_env_config)
        assert result.ips == ["172.17.0.1"]
        assert result.ports == [5601]
        assert result.execution_id == 10
        assert result.emulation_name == "test_emulation"



from typing import List, Dict, Any
from csle_common.dao.emulation_config.node_vulnerability_config import NodeVulnerabilityConfig
from csle_base.json_serializable import JSONSerializable


class VulnerabilitiesConfig(JSONSerializable):
    """
    A DTO class representing the vulnerabilities configuration of an emulation environment
    """

    def __init__(self, node_vulnerability_configs: List[NodeVulnerabilityConfig]):
        """
        Initializes the DTO

        :param node_vulnerability_configs: the list of Node vulnerability configurations
        """
        self.node_vulnerability_configs = node_vulnerability_configs

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d = {}
        d["vulnerabilities"] = list(map(lambda x: x.to_dict(), self.node_vulnerability_configs))
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "VulnerabilitiesConfig":
        """
        Converts a dict representation of the object to a DTO representation
        :return: a DTO representation of the object
        """
        vulnerabilities = list(map(lambda x: NodeVulnerabilityConfig.from_dict(x),
                                   d["vulnerabilities"]))
        dto = VulnerabilitiesConfig(node_vulnerability_configs=vulnerabilities)
        return dto

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return "vulnerabilities:{}".format(",".join(list(map(lambda x: str(x), self.node_vulnerability_configs))))

    def get_vulnerabilities(self, ips: List[str]) -> List[NodeVulnerabilityConfig]:
        """
        Gets a list of vulnerabilities for a list of ip addresses

        :param ips: the list of ip addresse
        :return: the list of vulnerabilities corresponding to the list of ip addresses
        """
        vulnerabilities = []
        for node_vuln_config in self.node_vulnerability_configs:
            if node_vuln_config.ip in ips:
                vulnerabilities.append(node_vuln_config)
        return vulnerabilities

    @staticmethod
    def from_json_file(json_file_path: str) -> "VulnerabilitiesConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return VulnerabilitiesConfig.from_dict(json.loads(json_str))

    def copy(self) -> "VulnerabilitiesConfig":
        """
        :return: a copy of the DTO
        """
        return VulnerabilitiesConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "VulnerabilitiesConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.node_vulnerability_configs = list(map(lambda x: x.create_execution_config(
            ip_first_octet=ip_first_octet), config.node_vulnerability_configs))
        return config



from typing import Dict, Any
from csle_base.json_serializable import JSONSerializable


class NmapHttpGrep(JSONSerializable):
    """
    A DTO representing the output of a Nmap HTTP Grep
    """

    def __init__(self, output: str):
        """
        Initializes the DTO

        :param output: the output of the HTTP Grep
        """
        self.output = output

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"output:{self.output}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d = {}
        d["output"] = self.output
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NmapHttpGrep":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NmapHttpGrep(output=d["output"])
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "NmapHttpGrep":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NmapHttpGrep.from_dict(json.loads(json_str))



from enum import IntEnum


class SystemModelType(IntEnum):
    """
    Enum representing the different system model types in CSLE
    """
    GAUSSIAN_MIXTURE = 0
    EMPIRICAL_DISTRIBUTION = 1
    GAUSSIAN_PROCESS = 2
    MCMC = 3



from unittest.mock import patch, MagicMock
from csle_common.util.connection_util import ConnectionUtil
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state import (
    EmulationConnectionObservationState,
)
import csle_common.constants.constants as constants


class TestConnectionUtilSuite:
    """
    Test suite for connection_util
    """

    @patch("paramiko.SSHClient")
    def test_ssh_setup_connection(self, mock_SSHClient) -> None:
        """
        Test the helper function for setting up a SSH connection

        :param mock_SSHClient: mock_SSHClient
        :return: None
        """
        mock_ssh_client = MagicMock()
        mock_SSHClient.return_value = mock_ssh_client
        mock_transport = MagicMock()
        mock_relay_channel = MagicMock()
        mock_ssh_client.get_transport.return_value = mock_transport
        mock_transport.open_channel.return_value = mock_relay_channel
        a = MagicMock()
        a.ips = ["192.168.1.2"]
        a.ips_match.side_effect = lambda ips: True
        credentials = [MagicMock()]
        credentials[0].service = constants.SSH.SERVICE_NAME
        credentials[0].username = "user"
        credentials[0].pw = "password"
        credentials[0].port = 22
        proxy_connections = [MagicMock()]
        proxy_connections[0].ip = "192.168.1.1"
        proxy_connections[0].conn.get_transport.return_value = mock_transport
        s = MagicMock()
        s.emulation_env_config.containers_config.agent_ip = "192.168.1.3"
        s.attacker_obs_state.agent_reachable = ["192.168.1.1"]
        result = ConnectionUtil._ssh_setup_connection(a=a, credentials=credentials,
                                                      proxy_connections=proxy_connections, s=s)
        mock_SSHClient.assert_not_called()
        assert result

    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_ssh_finalize_connection(self, mock_sleep, mock_execute_ssh_cmd) -> None:
        """
        Test the helper function for finalizing a SSH connection and setting up the DTO

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_sleep: mock_sleep
        :return: None
        """
        mock_execute_ssh_cmd.side_effect = [
            (b"output", b"error", 0.1),
            (b"output", b"error", 0.1),
            (b"output", b"error", 0.1),
            (b"output", b"error", 0.1),
            (b"(ALL) NOPASSWD: ALL", b"", 0.1),
        ]
        target_machine = MagicMock()
        connection_setup_dto = MagicMock()
        connection_setup_dto.target_connections = [MagicMock()]
        connection_setup_dto.credentials = [MagicMock()]
        connection_setup_dto.credentials[0].username = "user"
        connection_setup_dto.ports = [22]
        connection_setup_dto.proxies = [MagicMock()]
        connection_setup_dto.ip = "192.168.1.2"

        root, total_time = ConnectionUtil._ssh_finalize_connection(
            target_machine=target_machine, connection_setup_dto=connection_setup_dto, i=0
        )
        assert not root
        assert total_time
        mock_execute_ssh_cmd.assert_called()

    @patch("telnetlib.Telnet")
    @patch("time.sleep", return_value=None)
    def test_telnet_setup_connection(self, mock_sleep, mock_telnet) -> None:
        """
        Test the helper function for setting up a Telnet connection to a target machine

        :param mock_telnet: mock_telnet
        :param mock_sleep: mock_sleep
        :return: None
        """
        mock_telnet_conn = MagicMock()
        mock_telnet.return_value = mock_telnet_conn
        mock_telnet_conn.read_until.side_effect = [constants.TELNET.LOGIN_PROMPT, constants.TELNET.PASSWORD_PROMPT,
                                                   constants.TELNET.PROMPT]
        mock_telnet_conn.write.return_value = None
        mock_telnet_conn.read_until.return_value = b"$"
        a = MagicMock()
        a.ips = ["192.168.1.2"]
        a.ips_match.side_effect = lambda ips: True
        credentials = [MagicMock()]
        credentials[0].service = constants.TELNET.SERVICE_NAME
        credentials[0].username = "user"
        credentials[0].pw = "password"
        credentials[0].port = 23
        proxy_connections = [MagicMock()]
        proxy_connections[0].ip = "192.168.1.1"
        proxy_connections[0].conn.get_transport.return_value = MagicMock()
        s = MagicMock()
        s.emulation_env_config.containers_config.agent_ip = "192.168.1.3"
        s.attacker_obs_state.agent_reachable = ["192.168.1.1"]
        s.get_attacker_machine.return_value.reachable = ["192.168.1.2"]
        s.emulation_env_config.get_port_forward_port.return_value = 9999

        result = ConnectionUtil._telnet_setup_connection(
            a=a, credentials=credentials, proxy_connections=proxy_connections, s=s)
        mock_telnet.assert_called()
        assert result

    @patch("time.sleep", return_value=None)
    def test_telnet_finalize_connection(self, mock_sleep) -> None:
        """
        Test the helper function for finalizing a Telnet connection to a target machine and creating the DTO

        :param mock_sleep: mock_sleep
        :return: None
        """
        target_machine = MagicMock()
        connection_setup_dto = MagicMock()
        mock_telnet_conn = [MagicMock() for _ in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CHECK_ROOT)]
        connection_setup_dto.target_connections = mock_telnet_conn
        connection_setup_dto.credentials = [
            MagicMock() for _ in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CHECK_ROOT)]
        for credential in connection_setup_dto.credentials:
            credential.username = "user"
        connection_setup_dto.tunnel_threads = [
            MagicMock() for _ in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CHECK_ROOT)]
        connection_setup_dto.forward_ports = [9999 for _ in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CHECK_ROOT)]
        connection_setup_dto.ports = [23 for _ in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CHECK_ROOT)]
        connection_setup_dto.proxies = [MagicMock() for _ in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CHECK_ROOT)]
        connection_setup_dto.ip = "192.168.1.2"

        for i, conn in enumerate(mock_telnet_conn):
            conn.read_until.side_effect = [b"user may not run sudo" if i < 4 else b"(ALL) NOPASSWD: ALL"]

        root, total_time = ConnectionUtil._telnet_finalize_connection(
            target_machine=target_machine, i=0, connection_setup_dto=connection_setup_dto)
        assert not root
        assert total_time

    @patch("ftplib.FTP")
    def test_ftp_setup_connection(self, mock_ftp) -> None:
        """
        Test the helper function for setting up a FTP connection

        :param mock_ftp: mock_ftp
        :return: None
        """
        mock_ftp_conn = MagicMock()
        mock_ftp.return_value = mock_ftp_conn
        mock_ftp_conn.login.return_value = "230 Login successful."

        a = MagicMock()
        a.ips = ["192.168.1.2"]
        a.ips_match.side_effect = lambda ips: True

        credentials = [MagicMock()]
        credentials[0].service = constants.FTP.SERVICE_NAME
        credentials[0].username = "user"
        credentials[0].pw = "password"
        credentials[0].port = 21

        proxy_connections = [MagicMock()]
        proxy_connections[0].ip = "192.168.1.1"
        proxy_connections[0].conn.get_transport.return_value = MagicMock()
        proxy_connections[0].conn.invoke_shell.return_value = MagicMock()

        s = MagicMock()
        s.emulation_env_config.containers_config.agent_ip = "192.168.1.3"
        s.attacker_obs_state.agent_reachable = ["192.168.1.1"]
        s.get_attacker_machine.return_value.reachable = ["192.168.1.2"]
        s.emulation_env_config.get_port_forward_port.return_value = 9999

        result = ConnectionUtil._ftp_setup_connection(
            a=a, credentials=credentials, proxy_connections=proxy_connections, s=s)
        mock_ftp.assert_not_called()
        assert result

    def test_ftp_finalize_connection(self) -> None:
        """
        Test helper function for creating the connection DTO for FTP

        :return: None
        """
        target_machine = MagicMock()
        target_machine.ftp_connections = []

        connection_setup_dto = MagicMock()
        connection_setup_dto.target_connections = [MagicMock()]
        connection_setup_dto.credentials = [MagicMock()]
        connection_setup_dto.tunnel_threads = [MagicMock()]
        connection_setup_dto.forward_ports = [9999]
        connection_setup_dto.ports = [21]
        connection_setup_dto.interactive_shells = [MagicMock()]
        connection_setup_dto.proxies = [MagicMock()]
        connection_setup_dto.ip = "192.168.1.2"

        root, cost = ConnectionUtil._ftp_finalize_connection(
            target_machine=target_machine, i=0, connection_setup_dto=connection_setup_dto)
        assert not root
        assert cost == 0

    def test__find_jump_host_connection(self) -> None:
        """
        Test utility function for finding a jump-host from the set of compromised machines to reach a target IP

        :return: None
        """
        ip = "192.168.1.100"
        s = MagicMock()
        s.attacker_obs_state = MagicMock()
        s.attacker_obs_state.agent_reachable = ["192.168.1.100"]
        s.emulation_env_config.containers_config.agent_ip = "192.168.1.1"
        s.emulation_env_config.get_hacker_connection.return_value = MagicMock()

        result = ConnectionUtil.find_jump_host_connection(ip=ip, s=s)
        assert result.ip == "192.168.1.1"
        assert result.root

    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    def test_test_connection(self, mock_execute_ssh_cmd) -> None:
        """
        Test utility function for testing if a connection is alive or not

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :return: None
        """
        mock_conn = MagicMock()
        mock_execute_ssh_cmd.return_value = (b"user", b"", 0.1)
        c = EmulationConnectionObservationState(
            conn=mock_conn, credential=MagicMock(), root=False, service="ssh", port=22)
        result = ConnectionUtil.test_connection(c=c)
        assert result

    @patch("paramiko.SSHClient")
    def test_reconnect_ssh(self, mock_SSHClient) -> None:
        """
        Test the method that reconnects the given SSH connection if it has died for some reason

        :param mock_SSHClient: mock_SSHClient

        :return: None
        """
        mock_credential = MagicMock()
        mock_credential.username = "user"
        mock_credential.pw = "password"
        mock_credential.port = 22

        c = EmulationConnectionObservationState(
            conn=MagicMock(), credential=mock_credential, root=False, service="ssh", port=22, ip="192.168.1.2")
        mock_ssh_client = mock_SSHClient.return_value
        mock_transport = MagicMock()
        mock_ssh_client.get_transport.return_value = mock_transport

        result = ConnectionUtil.reconnect_ssh(c)
        assert result.conn == mock_ssh_client

    @patch("csle_common.util.connection_util.ConnectionUtil.reconnect_ssh")
    @patch("telnetlib.Telnet")
    def test_reconnect_telnet(self, mock_telnet, mock_reconnect_ssh) -> None:
        """
        Test the method that reconnects the given Telnet connection if it has died for some reason

        :param mock_telnet: mock_telnet
        :param mock_reconnect_ssh: mock_reconnect_ssh
        :return: None
        """
        mock_credential = MagicMock()
        mock_credential.username = "user"
        mock_credential.pw = "password"

        mock_proxy_conn = MagicMock()
        proxy = EmulationConnectionObservationState(
            conn=mock_proxy_conn, credential=mock_credential, root=False, service="ssh", port=22, ip="192.168.1.1")
        c = EmulationConnectionObservationState(
            conn=None, credential=mock_credential, root=False, service="telnet", port=23, ip="192.168.1.2", proxy=proxy)

        mock_telnet_conn = mock_telnet.return_value
        mock_telnet_conn.read_until.side_effect = [constants.TELNET.LOGIN_PROMPT, constants.TELNET.PASSWORD_PROMPT,
                                                   constants.TELNET.PROMPT]
        mock_reconnect_ssh.return_value = proxy
        result = ConnectionUtil.reconnect_telnet(c)
        assert result.conn == mock_telnet_conn



from typing import Dict, Any
from csle_base.json_serializable import JSONSerializable


class DockerStatsManagerConfig(JSONSerializable):
    """
    Represents the configuration of the docker stats managers in a CSLE emulation
    """

    def __init__(self, docker_stats_manager_log_file: str, docker_stats_manager_log_dir: str,
                 docker_stats_manager_max_workers: int,
                 time_step_len_seconds: int = 15, docker_stats_manager_port: int = 50046,
                 version: str = "0.0.1") -> None:
        """
        Initializes the DTO

        :param time_step_len_seconds: the length of a time-step (period for logging)
        :param version: the version
        :param docker_stats_manager_port: the GRPC port of the docker stats manager
        :param docker_stats_manager_log_file: log file of the docker stats manager
        :param docker_stats_manager_log_dir: log dir of the docker stats manager
        :param docker_stats_manager_max_workers: max number of gRPC workers of the docker stats manager
        """
        self.time_step_len_seconds = time_step_len_seconds
        self.version = version
        self.docker_stats_manager_port = docker_stats_manager_port
        self.docker_stats_manager_log_file = docker_stats_manager_log_file
        self.docker_stats_manager_log_dir = docker_stats_manager_log_dir
        self.docker_stats_manager_max_workers = docker_stats_manager_max_workers

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DockerStatsManagerConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = DockerStatsManagerConfig(time_step_len_seconds=d["time_step_len_seconds"],
                                       version=d["version"], docker_stats_manager_port=d["docker_stats_manager_port"],
                                       docker_stats_manager_log_file=d["docker_stats_manager_log_file"],
                                       docker_stats_manager_log_dir=d["docker_stats_manager_log_dir"],
                                       docker_stats_manager_max_workers=d["docker_stats_manager_max_workers"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["docker_stats_manager_port"] = self.docker_stats_manager_port
        d["time_step_len_seconds"] = self.time_step_len_seconds
        d["version"] = self.version
        d["docker_stats_manager_log_file"] = self.docker_stats_manager_log_file
        d["docker_stats_manager_log_dir"] = self.docker_stats_manager_log_dir
        d["docker_stats_manager_max_workers"] = self.docker_stats_manager_max_workers
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"docker_stats_manager_port: {self.docker_stats_manager_port}, " \
               f"time_step_len_seconds: {self.time_step_len_seconds}," \
               f" version: {self.version}, docker_stats_manager_log_dir: {self.docker_stats_manager_log_dir}, " \
               f"docker_stats_manager_log_file: {self.docker_stats_manager_log_file}, " \
               f"docker_stats_manager_max_workers: {self.docker_stats_manager_max_workers}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "DockerStatsManagerConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return DockerStatsManagerConfig.from_dict(json.loads(json_str))

    def copy(self) -> "DockerStatsManagerConfig":
        """
        :return: a copy of the DTO
        """
        return DockerStatsManagerConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "DockerStatsManagerConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        return config

    @staticmethod
    def schema() -> "DockerStatsManagerConfig":
        """
        :return: get the schema of the DTO
        """
        return DockerStatsManagerConfig(docker_stats_manager_log_file="docker_stats_manager.log",
                                        docker_stats_manager_log_dir="/var/log/csle/", docker_stats_manager_port=50046,
                                        docker_stats_manager_max_workers=10)



"""
Type of delay distributions on a container's network interface
"""
from enum import IntEnum


class PacketDelayDistributionType(IntEnum):
    """
    Enum representing the different types of delay distributions to emulate on a container
    """
    UNIFORM = 0
    NORMAL = 1
    PARETO = 2
    PARETONORMAL = 3



from typing import Dict, Any
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_base.json_serializable import JSONSerializable


class ExperimentExecution(JSONSerializable):
    """
    DTO representing an experiment execution
    """

    def __init__(self, config: ExperimentConfig, result: ExperimentResult, timestamp: float, emulation_name: str,
                 simulation_name: str, descr: str, log_file_path: str):
        """
        Initializes the DTO

        :param config: the experiment configuration
        :param result: the experiment result
        :param timestamp: the timestamp
        :param emulation_name: the emulation name
        :param simulation_name: the simulation name
        :param descr: a description of the training run
        """
        self.config = config
        self.result = result
        self.timestamp = timestamp
        self.emulation_name = emulation_name
        self.simulation_name = simulation_name
        self.id = -1
        self.descr = descr
        self.log_file_path = log_file_path

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ExperimentExecution":
        """
        Converts a dict representation of the object

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ExperimentExecution(
            config=ExperimentConfig.from_dict(d["config"]),
            result=ExperimentResult.from_dict(d["result"]),
            timestamp=d["timestamp"], simulation_name=d["simulation_name"], emulation_name=d["emulation_name"],
            descr=d["descr"], log_file_path=d["log_file_path"]
        )
        obj.id = d["id"]
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["config"] = self.config.to_dict()
        d["result"] = self.result.to_dict()
        d["timestamp"] = self.timestamp
        d["simulation_name"] = self.simulation_name
        d["emulation_name"] = self.emulation_name
        d["id"] = self.id
        d["descr"] = self.descr
        d["log_file_path"] = self.log_file_path
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"config: {self.config}, result: {self.result}, timestamp: {self.timestamp}, " \
               f"simulation_name: {self.simulation_name}, emulation_name: {self.emulation_name}, id: {self.id}," \
               f"descr: {self.descr}, log_file_path: {self.log_file_path}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ExperimentExecution":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ExperimentExecution.from_dict(json.loads(json_str))



from typing import Dict, Any
import numpy as np
from numpy.typing import NDArray
from csle_base.json_serializable import JSONSerializable


class ObservationFunctionConfig(JSONSerializable):
    """
    DTO representing the configuration of the observation function of a simulation
    """

    def __init__(self, observation_tensor: NDArray[Any], component_observation_tensors: Dict[str, NDArray[Any]]):
        """
        Initializes the DTO
        :param observation_tensor: the observation tensor
        """
        self.observation_tensor = observation_tensor
        self.component_observation_tensors = component_observation_tensors

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ObservationFunctionConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ObservationFunctionConfig(observation_tensor=d["observation_tensor"],
                                        component_observation_tensors=d["component_observation_tensors"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        tensor = list(self.observation_tensor)
        for i in range(len(tensor)):
            if isinstance(tensor[i], np.ndarray):
                tensor[i] = tensor[i].tolist()
        d["observation_tensor"] = list(tensor)
        d["component_observation_tensors"] = self.component_observation_tensors
        for k, v in d["component_observation_tensors"].items():
            if isinstance(v, np.ndarray):
                d["component_observation_tensors"][k] = v.tolist()
            for i in range(len(d["component_observation_tensors"][k])):
                if isinstance(d["component_observation_tensors"][k][i], np.ndarray):
                    d["component_observation_tensors"][k][i] = d["component_observation_tensors"][k][i].tolist()
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"observation tensor: {self.observation_tensor}, " \
               f"component_observation_tensors: {self.component_observation_tensors}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ObservationFunctionConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ObservationFunctionConfig.from_dict(json.loads(json_str))



import time
from typing import Dict, Any, Union
from csle_base.json_serializable import JSONSerializable


class SessionToken(JSONSerializable):
    """
    DTO representing a session token
    """

    def __init__(self, token: str, timestamp: float, username: str) -> None:
        """
        Initializes the DTO

        :param token: the token string
        :param timestamp: the time the token was created
        :param username: the username of the user that the token is given to
        """
        self.token = token
        self.timestamp = timestamp
        self.username = username

    def to_dict(self) -> Dict[str, Union[str, float]]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, float]] = {}
        d["username"] = self.username
        d["timestamp"] = self.timestamp
        d["token"] = self.token
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SessionToken":
        """
        Converts a dict representation of the DTO into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = SessionToken(
            username=d["username"], token=d["token"], timestamp=d["timestamp"]
        )
        return obj

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"username: {self.username}, token: {self.token}, timestamp: {self.timestamp}"

    def expired(self, valid_length_hours: int) -> bool:
        """
        Checks if the token has expired or not

        :param valid_length_hours: the number of hours that a token is valid
        :return: True if the token has expired otherwise False
        """
        return ((time.time() - self.timestamp) / (60) / (60)) > valid_length_hours

    @staticmethod
    def from_json_file(json_file_path: str) -> "SessionToken":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return SessionToken.from_dict(json.loads(json_str))



from typing import List, Dict, Any
from csle_common.dao.simulation_config.action_space_config import ActionSpaceConfig
from csle_base.json_serializable import JSONSerializable


class JointActionSpaceConfig(JSONSerializable):
    """
    DTO representing the joint action space of a simulation environment
    """

    def __init__(self, action_spaces: List[ActionSpaceConfig]):
        """
        Initializes the DTO

        :param action_spaces: the list of action spaces for each player
        """
        self.action_spaces = action_spaces

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "JointActionSpaceConfig":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = JointActionSpaceConfig(
            action_spaces=list(map(lambda x: ActionSpaceConfig.from_dict(x), d["action_spaces"]))
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d = {}
        d["action_spaces"] = list(map(lambda x: x.to_dict(), self.action_spaces))
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"action_spaces: {list(map(lambda x: str(x), self.action_spaces))}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "JointActionSpaceConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return JointActionSpaceConfig.from_dict(json.loads(json_str))



from typing import List, Dict, Any, Union
from csle_base.json_serializable import JSONSerializable


class KafkaTopic(JSONSerializable):
    """
    DTO representing a kafka topic (Records are assumed to be comma-separated strings) where attributes define
    the list of columns in the csv.
    """

    def __init__(self, name: str, num_partitions: int, num_replicas: int, attributes: List[str],
                 retention_time_hours: int):
        """
        Initializes the DTO

        :param name: the name of the topic
        :param num_partitions: the number of partitions
        :param num_replicas: the number of replicas
        :param attributes: the attributes of the topic
        :param retention_time_hours: the retention time of the topic (how long to store the data)
        """
        self.name = name
        self.num_partitions = num_partitions
        self.num_replicas = num_replicas
        self.attributes = attributes
        self.retention_time_hours = retention_time_hours

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "KafkaTopic":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = KafkaTopic(name=d["name"], num_replicas=d["num_replicas"], num_partitions=d["num_partitions"],
                         attributes=d["attributes"], retention_time_hours=d["retention_time_hours"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, int, List[str]]] = {}
        d["name"] = self.name
        d["num_partitions"] = self.num_partitions
        d["num_replicas"] = self.num_replicas
        d["attributes"] = self.attributes
        d["retention_time_hours"] = self.retention_time_hours
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"name:{self.name}, num_partitions: {self.num_partitions}, " \
               f"num_replicas:{self.num_replicas}, attributes: {','.join(self.attributes)}, " \
               f"retention_time_hours:{self.retention_time_hours}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "KafkaTopic":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return KafkaTopic.from_dict(json.loads(json_str))

    def copy(self) -> "KafkaTopic":
        """
        :return: a copy of the DTO
        """
        return KafkaTopic.from_dict(self.to_dict())

    @staticmethod
    def schema() -> "KafkaTopic":
        """
        :return: get the schema of the DTO
        """
        return KafkaTopic(name="", num_partitions=1, num_replicas=1, attributes=[""], retention_time_hours=1)



from typing import List, Dict, Any
from csle_common.dao.emulation_config.network_service import NetworkService
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class NodeServicesConfig(JSONSerializable):
    """
    A DTO object representing the services configuration of a node in an emulation environment
    """

    def __init__(self, ip: str, services: List[NetworkService]):
        """
        Initializes the DTO

        :param users: the list of node users configuration
        """
        self.ip = ip
        self.services = services

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["services"] = list(map(lambda x: x.to_dict(), self.services))
        d["ip"] = self.ip
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NodeServicesConfig":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = NodeServicesConfig(ip=d["ip"],
                                 services=list(map(lambda x: NetworkService.from_dict(x), d["services"])))
        return dto

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return "ip:{}, services:{}".format(self.ip, ",".join(list(map(lambda x: str(x), self.services))))

    @staticmethod
    def from_json_file(json_file_path: str) -> "NodeServicesConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NodeServicesConfig.from_dict(json.loads(json_str))

    def copy(self) -> "NodeServicesConfig":
        """
        :return: a copy of the DTO
        """
        return NodeServicesConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "NodeServicesConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.ip = GeneralUtil.replace_first_octet_of_ip(ip=config.ip, ip_first_octet=ip_first_octet)
        return config



from typing import Union, Dict, Any
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_observation.attacker.emulation_attacker_observation_state \
    import EmulationAttackerObservationState
from csle_common.dao.emulation_observation.defender.emulation_defender_observation_state \
    import EmulationDefenderObservationState
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
from csle_common.dao.emulation_observation.defender.emulation_defender_machine_observation_state \
    import EmulationDefenderMachineObservationState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_config import EmulationAttackerActionConfig
from csle_common.dao.emulation_action.defender.emulation_defender_action_config import EmulationDefenderActionConfig
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state import \
    EmulationConnectionObservationState
from csle_common.dao.emulation_config.credential import Credential
from csle_base.json_serializable import JSONSerializable
from csle_collector.client_manager.client_population_metrics import ClientPopulationMetrics
from csle_collector.docker_stats_manager.dao.docker_stats import DockerStats
from csle_collector.snort_ids_manager.dao.snort_ids_alert_counters import SnortIdsAlertCounters
from csle_collector.snort_ids_manager.dao.snort_ids_rule_counters import SnortIdsRuleCounters
from csle_collector.ossec_ids_manager.dao.ossec_ids_alert_counters import OSSECIdsAlertCounters
from csle_collector.host_manager.dao.host_metrics import HostMetrics


class EmulationEnvState(JSONSerializable):
    """
    Represents the combined state of the emulation environment,
    including both the attacker's and the defender's states.
    """

    def __init__(self, emulation_env_config: EmulationEnvConfig):
        """
        Initializes the state

        :param emulation_env_config: the environment configuration
        :param attacker_action_config: the configuration of the attacker agent
        :param defender_agent_config: the configuration of the defender agent
        """
        self.emulation_env_config = emulation_env_config
        self.attacker_action_config = EmulationAttackerActionConfig.all_actions_config(
            num_nodes=len(self.emulation_env_config.containers_config.containers),
            subnet_masks=self.emulation_env_config.topology_config.subnetwork_masks,
            hacker_ip=self.emulation_env_config.containers_config.agent_ip)
        self.defender_action_config = EmulationDefenderActionConfig.all_actions_config(
            num_nodes=len(self.emulation_env_config.containers_config.containers),
            subnet_masks=self.emulation_env_config.topology_config.subnetwork_masks)
        self.vuln_lookup = constants.VULNERABILITIES.vuln_lookup
        self.vuln_lookup_inv = {v: k for k, v in self.vuln_lookup.items()}
        self.service_lookup = constants.SERVICES.service_lookup
        self.service_lookup_inv = {v: k for k, v in self.service_lookup.items()}
        self.os_lookup = constants.OS.os_lookup
        self.os_lookup_inv = {v: k for k, v in self.os_lookup.items()}
        self.attacker_obs_state: Union[EmulationAttackerObservationState, None] = None
        self.defender_obs_state: Union[EmulationDefenderObservationState, None] = None
        self.attacker_cached_ssh_connections: Dict[Any, Any] = {}
        self.attacker_cached_telnet_connections: Dict[Any, EmulationConnectionObservationState] = {}
        self.attacker_cached_ftp_connections: Dict[Any, EmulationConnectionObservationState] = {}
        self.defender_cached_ssh_connections: Dict[Any, EmulationConnectionObservationState] = {}
        self.attacker_cached_backdoor_credentials: Dict[Any, Credential] = {}
        self.reset()

    def initialize_defender_machines(self) -> None:
        """
        Initializes the defender observation state based on the emulation configuration

        :return: None
        """
        defender_machines = []
        for c in self.emulation_env_config.containers_config.containers:
            defender_machines.append(EmulationDefenderMachineObservationState.from_container(
                c, kafka_config=self.emulation_env_config.kafka_config))
        if self.defender_obs_state is None:
            raise ValueError("EmulationDefenderObservationState is None")
        self.defender_obs_state.machines = defender_machines
        self.defender_obs_state.start_monitoring_threads()

    def reset(self) -> None:
        """
        Resets the env state. Caches connections

        :return: None
        """
        agent_reachable = set()
        if self.attacker_obs_state is not None:
            agent_reachable = self.attacker_obs_state.agent_reachable
            for m in self.attacker_obs_state.machines:
                for c in m.ssh_connections:
                    self.attacker_cached_ssh_connections[(c.ip, c.credential.username, c.port)] = c
                for c in m.telnet_connections:
                    self.attacker_cached_telnet_connections[(c.ip, c.credential.username, c.port)] = c
                for c in m.ftp_connections:
                    self.attacker_cached_ftp_connections[(c.ip, c.credential.username, c.port)] = c
                for cr in m.backdoor_credentials:
                    for ip in m.ips:
                        self.attacker_cached_backdoor_credentials[(ip, cr.username, cr.pw)] = cr
        self.attacker_obs_state = EmulationAttackerObservationState(catched_flags=0, agent_reachable=agent_reachable)
        self.attacker_obs_state.last_attacker_action = None
        self.attacker_obs_state.undetected_intrusions_steps = 0
        self.attacker_obs_state.all_flags = False
        self.attacker_obs_state.catched_flags = 0
        self.attacker_obs_state.step = 1

        if self.defender_obs_state is not None:
            for m in self.defender_obs_state.machines:
                for c in m.ssh_connections:
                    self.defender_cached_ssh_connections[(c.ip, c.credential.username, c.port)] = c
        else:
            self.defender_obs_state = EmulationDefenderObservationState(
                kafka_config=self.emulation_env_config.kafka_config,
                client_population_metrics=ClientPopulationMetrics(),
                docker_stats=DockerStats(), snort_ids_alert_counters=SnortIdsAlertCounters(),
                snort_ids_rule_counters=SnortIdsRuleCounters(), ossec_ids_alert_counters=OSSECIdsAlertCounters(),
                aggregated_host_metrics=HostMetrics(), defender_actions=[], attacker_actions=[])

    def cleanup(self) -> None:
        """
        Cleanup

        :return: None
        """
        for _, c in self.attacker_cached_ssh_connections.items():
            c.cleanup()
        for _, c in self.attacker_cached_ftp_connections.items():
            c.cleanup()
        for _, c in self.attacker_cached_telnet_connections.items():
            c.cleanup()

        if self.attacker_obs_state is None:
            raise ValueError("EmualtionAttackerObservationState is None")
        self.attacker_obs_state.cleanup()

        for _, c in self.defender_cached_ssh_connections.items():
            (ssh_conns, _) = c
            for c2 in ssh_conns:
                c2.cleanup()
        if self.defender_obs_state is None:
            raise ValueError("EmulationDefenderObservationState is None")
        self.defender_obs_state.cleanup()

    def get_attacker_machine(self, ip: str) -> Union[EmulationAttackerMachineObservationState, None]:
        """
        Utility function for extracting the attacker machine from the attacker's observation

        :param ip: the ip of the attacker machine
        :return: the machine if is found, otherwise None
        """
        if self.attacker_obs_state is None:
            raise ValueError("EmulationAttackerObservationState is None")
        for m in self.attacker_obs_state.machines:
            for m_ip in m.ips:
                if m_ip == ip:
                    return m
        return None

    def get_defender_machine(self, ip: str) -> Union[EmulationDefenderMachineObservationState, None]:
        """
        Utility function for extracting the defender machine from the defender's observation given an IP

        :param ip: the ip of the machine
        :return: the machine if found otherwise None
        """
        if self.defender_obs_state is None:
            raise ValueError("EmulationDefenderObservationState is None")
        for m in self.defender_obs_state.machines:
            for m_ip in m.ips:
                if m_ip == ip:
                    return m
        return None

    def copy(self) -> "EmulationEnvState":
        """
        :return: a copy of the env state
        """
        copy = EmulationEnvState(emulation_env_config=self.emulation_env_config)
        if self.attacker_obs_state is None:
            raise ValueError("EmulationAttackerObservationState is None")
        if self.defender_obs_state is None:
            raise ValueError("EmulationDefenderObservationState is None")
        copy.attacker_obs_state = self.attacker_obs_state.copy()
        copy.defender_obs_state = self.defender_obs_state.copy()
        return copy

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"Attacker observation state: {self.attacker_obs_state}" \
               f"Defender observation state: {self.defender_obs_state}"

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationEnvState":
        """
        Converts a dict representation of the object into a an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationEnvState(emulation_env_config=EmulationEnvConfig.from_dict(d["emulation_env_config"]))
        obj.attacker_action_config = EmulationAttackerActionConfig.from_dict(d["attacker_action_config"])
        obj.defender_action_config = EmulationDefenderActionConfig.from_dict(d["defender_action_config"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        
        :return: a dict representation of the object
        """
        d = {}
        d["emulation_env_config"] = self.emulation_env_config.to_dict()
        d["attacker_action_config"] = self.attacker_action_config.to_dict()
        d["defender_action_config"] = self.defender_action_config.to_dict()
        return d

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationEnvState":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationEnvState.from_dict(json.loads(json_str))
